from galaxy.util.dictifiable import Dictifiable
from galaxy.util.odict import odict


class ToolOutputBase( Dictifiable, object ):

    def __init__( self, name, label=None, filters=None, hidden=False ):
        super( ToolOutputBase, self ).__init__()
        self.name = name
        self.label = label
        self.filters = filters or []
        self.hidden = hidden
        self.collection = False


class ToolOutput( ToolOutputBase ):
    """
    Represents an output datasets produced by a tool. For backward
    compatibility this behaves as if it were the tuple::

      (format, metadata_source, parent)
    """

    dict_collection_visible_keys = ( 'name', 'format', 'label', 'hidden' )

    def __init__( self, name, format=None, format_source=None, metadata_source=None,
                  parent=None, label=None, filters=None, actions=None, hidden=False,
                  implicit=False ):
        super( ToolOutput, self ).__init__( name, label=label, filters=filters, hidden=hidden )
        self.format = format
        self.format_source = format_source
        self.metadata_source = metadata_source
        self.parent = parent
        self.actions = actions

        # Initialize default values
        self.change_format = []
        self.implicit = implicit
        self.from_work_dir = None

    # Tuple emulation

    def __len__( self ):
        return 3

    def __getitem__( self, index ):
        if index == 0:
            return self.format
        elif index == 1:
            return self.metadata_source
        elif index == 2:
            return self.parent
        else:
            raise IndexError( index )

    def __iter__( self ):
        return iter( ( self.format, self.metadata_source, self.parent ) )

    def to_dict( self, view='collection', value_mapper=None, app=None ):
        as_dict = super( ToolOutput, self ).to_dict( view=view, value_mapper=value_mapper )
        format = self.format
        if format and format != "input" and app:
            edam_format = app.datatypes_registry.edam_formats.get(self.format)
            as_dict["edam_format"] = edam_format
            edam_data = app.datatypes_registry.edam_data.get(self.format)
            as_dict["edam_data"] = edam_data
        return as_dict


class ToolOutputCollection( ToolOutputBase ):
    """
    Represents a HistoryDatasetCollectionAssociation of output datasets produced
    by a tool.

    <outputs>
      <collection type="list" label="${tool.name} on ${on_string} fasta">
        <discover_datasets pattern="__name__" ext="fasta" visible="True" directory="outputFiles" />
      </collection>
      <collection type="paired" label="${tool.name} on ${on_string} paired reads">
        <data name="forward" format="fastqsanger" />
        <data name="reverse" format="fastqsanger"/>
      </collection>
    <outputs>
    """

    def __init__(
        self,
        name,
        structure,
        label=None,
        filters=None,
        hidden=False,
        default_format="data",
        default_format_source=None,
        default_metadata_source=None,
        inherit_format=False,
        inherit_metadata=False
    ):
        super( ToolOutputCollection, self ).__init__( name, label=label, filters=filters, hidden=hidden )
        self.collection = True
        self.default_format = default_format
        self.structure = structure
        self.outputs = odict()

        self.inherit_format = inherit_format
        self.inherit_metadata = inherit_metadata

        self.metadata_source = default_metadata_source
        self.format_source = default_format_source
        self.change_format = []  # TODO

    def known_outputs( self, inputs, type_registry ):
        if self.dynamic_structure:
            return []

        # This line is probably not right - should verify structured_like
        # or have outputs and all outputs have name.
        if len( self.outputs ) > 1:
            output_parts = [ToolOutputCollectionPart(self, k, v) for k, v in self.outputs.items()]
        else:
            # either must have specified structured_like or something worse
            if self.structure.structured_like:
                collection_prototype = inputs[ self.structure.structured_like ].collection
            else:
                collection_prototype = type_registry.prototype( self.structure.collection_type )

            def prototype_dataset_element_to_output( element, parent_ids=[] ):
                name = element.element_identifier
                format = self.default_format
                if self.inherit_format:
                    format = element.dataset_instance.ext
                output = ToolOutput(
                    name,
                    format=format,
                    format_source=self.format_source,
                    metadata_source=self.metadata_source,
                    implicit=True,
                )
                if self.inherit_metadata:
                    output.metadata_source = element.dataset_instance
                return ToolOutputCollectionPart(
                    self,
                    element.element_identifier,
                    output,
                    parent_ids=parent_ids,
                )

            def prototype_collection_to_output( collection_prototype, parent_ids=[] ):
                output_parts = []
                for element in collection_prototype.elements:
                    element_parts = []
                    if not element.is_collection:
                        element_parts.append(prototype_dataset_element_to_output( element, parent_ids ))
                    else:
                        new_parent_ids = parent_ids[:] + [element.element_identifier]
                        element_parts.extend(prototype_collection_to_output(element.element_object, new_parent_ids))
                    output_parts.extend(element_parts)

                return output_parts

            output_parts = prototype_collection_to_output( collection_prototype )

        return output_parts

    @property
    def dynamic_structure(self):
        return self.structure.dynamic

    @property
    def dataset_collector_descriptions(self):
        if not self.dynamic_structure:
            raise Exception("dataset_collector_descriptions called for output collection with static structure")
        return self.structure.dataset_collector_descriptions


class ToolOutputCollectionStructure( object ):

    def __init__(
        self,
        collection_type,
        collection_type_source,
        structured_like,
        dataset_collector_descriptions,
    ):
        self.collection_type = collection_type
        self.collection_type_source = collection_type_source
        self.structured_like = structured_like
        self.dataset_collector_descriptions = dataset_collector_descriptions
        if collection_type and collection_type_source:
            raise ValueError("Cannot set both type and type_source on collection output.")
        if collection_type is None and structured_like is None and dataset_collector_descriptions is None and collection_type_source is None:
            raise ValueError( "Output collection types must be specify type of structured_like" )
        if dataset_collector_descriptions and structured_like:
            raise ValueError( "Cannot specify dynamic structure (discovered_datasets) and structured_like attribute." )
        self.dynamic = dataset_collector_descriptions is not None


class ToolOutputCollectionPart( object ):

    def __init__( self, output_collection_def, element_identifier, output_def, parent_ids=[] ):
        self.output_collection_def = output_collection_def
        self.element_identifier = element_identifier
        self.output_def = output_def
        self.parent_ids = parent_ids

    @property
    def effective_output_name( self ):
        name = self.output_collection_def.name
        part_name = self.element_identifier
        effective_output_name = "%s|__part__|%s" % ( name, part_name )
        return effective_output_name

    @staticmethod
    def is_named_collection_part_name( name ):
        return "|__part__|" in name

    @staticmethod
    def split_output_name( name ):
        assert ToolOutputCollectionPart.is_named_collection_part_name( name )
        return name.split("|__part__|")



import functools
import operator

from galaxy import util


# Utility methods for specifing maps.
def to_str_or_none( value ):
    if value is None:
        return None
    else:
        return str( value )


def to_bool_or_none( value ):
    return util.string_as_bool_or_none( value )


def to_bool( value ):
    return util.asbool( value )


def to_float_or_none( value ):
    if value is None:
        return None
    else:
        return float( value )


# Utility methods for specifing valid...
def is_in( *args ):
    return functools.partial( operator.contains, args )



"""This module contains a linting function for a tool's command description.

A command description describes how to build the command-line to execute
from supplied inputs.
"""


def lint_command(tool_xml, lint_ctx):
    """Ensure tool contains exactly one command and check attributes."""
    root = tool_xml.getroot()
    commands = root.findall("command")
    if len(commands) > 1:
        lint_ctx.error("More than one command tag found, behavior undefined.")
        return

    if len(commands) == 0:
        lint_ctx.error("No command tag found, must specify a command template to execute.")
        return

    command = get_command(tool_xml)
    if "TODO" in command:
        lint_ctx.warn("Command template contains TODO text.")

    command_attrib = command.attrib
    interpreter_type = None
    for key, value in command_attrib.items():
        if key == "interpreter":
            interpreter_type = value
        elif key == "detect_errors":
            detect_errors = value
            if detect_errors not in ["default", "exit_code", "aggressive"]:
                lint_ctx.warn("Unknown detect_errors attribute [%s]" % detect_errors)
        else:
            lint_ctx.warn("Unknown attribute [%s] encountered on command tag." % key)

    interpreter_info = ""
    if interpreter_type:
        interpreter_info = " with interpreter of type [%s]" % interpreter_type
    if interpreter_type:
        lint_ctx.info("Command uses deprecated 'interpreter' attribute.")
    lint_ctx.info("Tool contains a command%s." % interpreter_info)


def get_command(tool_xml):
    """Get command XML element from supplied XML root."""
    root = tool_xml.getroot()
    commands = root.findall("command")
    command = None
    if len(commands) == 1:
        command = commands[0]
    return command



import logging

from os import listdir
from os.path import (
    basename,
    exists,
    isdir,
    islink,
    join,
    realpath,
)

from .resolver_mixins import UsesToolDependencyDirMixin

from ..resolvers import (
    Dependency,
    DependencyResolver,
    ListableDependencyResolver,
    MappableDependencyResolver,
    NullDependency,
)

log = logging.getLogger( __name__ )


class GalaxyPackageDependency(Dependency):
    dict_collection_visible_keys = Dependency.dict_collection_visible_keys + ['script', 'path', 'version', 'name']
    dependency_type = 'galaxy_package'

    def __init__( self, script, path, version, name, exact=True ):
        self.script = script
        self.path = path
        self.version = version
        self.name = name
        self._exact = exact

    @property
    def exact(self):
        return self._exact

    def shell_commands( self, requirement ):
        base_path = self.path
        if self.script is None and base_path is None:
            log.warning( "Failed to resolve dependency on '%s', ignoring", requirement.name )
            commands = None
        elif requirement.type == 'package' and self.script is None:
            commands = 'PACKAGE_BASE=%s; export PACKAGE_BASE; PATH="%s/bin:$PATH"; export PATH' % ( base_path, base_path )
        else:
            commands = 'PACKAGE_BASE=%s; export PACKAGE_BASE; . %s' % ( base_path, self.script )
        return commands


class ToolShedDependency(GalaxyPackageDependency):
    dependency_type = 'tool_shed_package'


class BaseGalaxyPackageDependencyResolver(DependencyResolver, UsesToolDependencyDirMixin):
    dict_collection_visible_keys = DependencyResolver.dict_collection_visible_keys + ['base_path', 'versionless']
    dependency_type = GalaxyPackageDependency

    def __init__(self, dependency_manager, **kwds):
        # Galaxy tool shed requires explicit versions on XML elements,
        # this in inconvient for testing or Galaxy instances not utilizing
        # the tool shed so allow a fallback version of the Galaxy package
        # resolver that will just grab 'default' version of exact version
        # unavailable.
        self.versionless = str(kwds.get('versionless', "false")).lower() == "true"
        self._init_base_path( dependency_manager, **kwds )

    def resolve(self, requirement, **kwds):
        """
        Attempt to find a dependency named `name` at version `version`. If version is None, return the "default" version as determined using a
        symbolic link (if found). Returns a triple of: env_script, base_path, real_version
        """
        name, version, type = requirement.name, requirement.version, requirement.type

        if version is None or self.versionless:
            exact = not self.versionless or version is None
            return self._find_dep_default( name, type=type, exact=exact, **kwds )
        else:
            return self._find_dep_versioned( name, version, type=type, **kwds )

    def _find_dep_versioned( self, name, version, type='package', **kwds ):
        base_path = self.base_path
        path = join( base_path, name, version )
        return self._galaxy_package_dep(path, version, name, True)

    def _find_dep_default( self, name, type='package', exact=True, **kwds ):
        base_path = self.base_path
        path = join( base_path, name, 'default' )
        if islink( path ):
            real_path = realpath( path )
            real_version = basename( real_path )
            return self._galaxy_package_dep(real_path, real_version, name, exact)
        else:
            return NullDependency(version=None, name=name)

    def _galaxy_package_dep( self, path, version, name, exact ):
        script = join( path, 'env.sh' )
        if exists( script ):
            return self.dependency_type(script, path, version, name, exact)
        elif exists( join( path, 'bin' ) ):
            return self.dependency_type(None, path, version, name, exact)
        return NullDependency(version=version, name=name)


class GalaxyPackageDependencyResolver(BaseGalaxyPackageDependencyResolver, ListableDependencyResolver, MappableDependencyResolver):
    resolver_type = "galaxy_packages"

    def __init__(self, dependency_manager, **kwds):
        super(GalaxyPackageDependencyResolver, self).__init__(dependency_manager, **kwds)
        self._setup_mapping(dependency_manager, **kwds)

    def resolve(self, requirement, **kwds):
        requirement = self._expand_mappings(requirement)
        return super(GalaxyPackageDependencyResolver, self).resolve(requirement, **kwds)

    def list_dependencies(self):
        base_path = self.base_path
        for package_name in listdir(base_path):
            package_dir = join(base_path, package_name)
            if isdir(package_dir):
                for version in listdir(package_dir):
                    version_dir = join(package_dir, version)
                    if version == "default":
                        version = None
                    valid_dependency = _is_dependency_directory(version_dir)
                    if valid_dependency:
                        yield self._to_requirement(package_name, version)


def _is_dependency_directory(directory):
    return exists(join(directory, 'env.sh')) or exists(join(directory, 'bin'))


__all__ = (
    'GalaxyPackageDependency',
    'GalaxyPackageDependencyResolver',
    'ToolShedDependency'
)



"""The module describes the ``core`` job metrics plugin."""
import logging
import time

from ..instrumenters import InstrumentPlugin
from ...metrics import formatting

log = logging.getLogger( __name__ )

GALAXY_SLOTS_KEY = "galaxy_slots"
START_EPOCH_KEY = "start_epoch"
END_EPOCH_KEY = "end_epoch"
RUNTIME_SECONDS_KEY = "runtime_seconds"


class CorePluginFormatter( formatting.JobMetricFormatter ):

    def format( self, key, value ):
        value = int( value )
        if key == GALAXY_SLOTS_KEY:
            return ( "Cores Allocated", "%d" % value )
        elif key == RUNTIME_SECONDS_KEY:
            return ( "Job Runtime (Wall Clock)", formatting.seconds_to_str( value ) )
        else:
            # TODO: Use localized version of this from galaxy.ini
            title = "Job Start Time" if key == START_EPOCH_KEY else "Job End Time"
            return (title, time.strftime( '%Y-%m-%d %H:%M:%S', time.localtime( value ) ) )


class CorePlugin( InstrumentPlugin ):
    """ Simple plugin that collects data without external dependencies. In
    particular it currently collects value set for Galaxy slots.
    """
    plugin_type = "core"
    formatter = CorePluginFormatter()

    def __init__( self, **kwargs ):
        pass

    def pre_execute_instrument( self, job_directory ):
        commands = []
        commands.append( self.__record_galaxy_slots_command( job_directory ) )
        commands.append( self.__record_seconds_since_epoch_to_file( job_directory, "start" ) )
        return commands

    def post_execute_instrument( self, job_directory ):
        commands = []
        commands.append( self.__record_seconds_since_epoch_to_file( job_directory, "end" ) )
        return commands

    def job_properties( self, job_id, job_directory ):
        galaxy_slots_file = self.__galaxy_slots_file( job_directory )

        properties = {}
        properties[ GALAXY_SLOTS_KEY ] = self.__read_integer( galaxy_slots_file )
        start = self.__read_seconds_since_epoch( job_directory, "start" )
        end = self.__read_seconds_since_epoch( job_directory, "end" )
        if start is not None and end is not None:
            properties[ START_EPOCH_KEY ] = start
            properties[ END_EPOCH_KEY ] = end
            properties[ RUNTIME_SECONDS_KEY ] = end - start
        return properties

    def __record_galaxy_slots_command( self, job_directory ):
        galaxy_slots_file = self.__galaxy_slots_file( job_directory )
        return '''echo "$GALAXY_SLOTS" > '%s' ''' % galaxy_slots_file

    def __record_seconds_since_epoch_to_file( self, job_directory, name ):
        path = self._instrument_file_path( job_directory, "epoch_%s" % name )
        return 'date +"%s" > ' + path

    def __read_seconds_since_epoch( self, job_directory, name ):
        path = self._instrument_file_path( job_directory, "epoch_%s" % name )
        return self.__read_integer( path )

    def __galaxy_slots_file( self, job_directory ):
        return self._instrument_file_path( job_directory, "galaxy_slots" )

    def __read_integer( self, path ):
        value = None
        try:
            value = int( open( path, "r" ).read() )
        except Exception:
            pass
        return value


__all__ = ( 'CorePlugin', )



"""The module describes the ``uname`` job metrics plugin."""
from ..instrumenters import InstrumentPlugin
from ...metrics import formatting


class UnameFormatter( formatting.JobMetricFormatter ):

    def format( self, key, value ):
        return "Operating System", value


class UnamePlugin( InstrumentPlugin ):
    """ Use uname to gather operating system information about remote system
    job is running on. Linux only.
    """
    plugin_type = "uname"
    formatter = UnameFormatter()

    def __init__( self, **kwargs ):
        self.uname_args = kwargs.get( "args", "-a" )

    def pre_execute_instrument( self, job_directory ):
        return "uname %s > '%s'" % ( self.uname_args, self.__instrument_uname_path( job_directory ) )

    def job_properties( self, job_id, job_directory ):
        properties = {}
        with open( self.__instrument_uname_path( job_directory ) ) as f:
            properties[ "uname" ] = f.read()
        return properties

    def __instrument_uname_path( self, job_directory ):
        return self._instrument_file_path( job_directory, "uname" )


__all__ = ( 'UnamePlugin', )



import logging
import sys

from copy import deepcopy

from galaxy.util import listify

log = logging.getLogger( __name__ )


class FilterFactory( object ):
    """
    An instance of this class is responsible for filtering the list
    of tools presented to a given user in a given context.
    """

    def __init__( self, toolbox ):
        self.toolbox = toolbox

        # Prepopulate dict containing filters that are always checked,
        # other filters that get checked depending on context (e.g. coming from
        # trackster or no user found are added in build filters).
        self.default_filters = dict( tool=[ _not_hidden, _handle_authorization ], section=[], label=[] )
        # Add dynamic filters to these default filters.
        config = toolbox.app.config
        self.__base_modules = listify( getattr( config, "toolbox_filter_base_modules", "galaxy.tools.filters" ) )
        self.__init_filters( "tool", getattr( config, "tool_filters", "" ), self.default_filters )
        self.__init_filters( "section", getattr( config, "tool_section_filters", "" ), self.default_filters )
        self.__init_filters( "label", getattr( config, "tool_label_filters", "" ), self.default_filters )

    def build_filters( self, trans, **kwds ):
        """
        Build list of filters to check tools against given current context.
        """
        filters = deepcopy( self.default_filters )
        if trans.user:
            for name, value in trans.user.preferences.items():
                if value.strip():
                    user_filters = listify( value, do_strip=True )
                    category = ''
                    if name == 'toolbox_tool_filters':
                        category = "tool"
                    elif name == 'toolbox_section_filters':
                        category = "section"
                    elif name == 'toolbox_label_filters':
                        category = "label"
                    if category:
                        validate = getattr( trans.app.config, 'user_%s_filters' % category, [] )
                        self.__init_filters( category, user_filters, filters, validate=validate )
        else:
            if kwds.get( "trackster", False ):
                filters[ "tool" ].append( _has_trackster_conf )

        return filters

    def __init_filters( self, key, filters, toolbox_filters, validate=None ):
        for filter in filters:
            if validate is None or filter in validate or filter in self.default_filters:
                filter_function = self.build_filter_function( filter )
                toolbox_filters[ key ].append( filter_function )
            else:
                log.warning( "Refusing to load %s filter '%s' which is not defined in config", key, filter )
        return toolbox_filters

    def build_filter_function( self, filter_name ):
        """Obtain python function (importing a submodule if needed)
        corresponding to filter_name.
        """
        if ":" in filter_name:
            # Should be a submodule of filters (e.g. examples:restrict_development_tools)
            (module_name, function_name) = filter_name.rsplit(":", 1)
            function = self._import_filter( module_name, function_name )
        else:
            # No module found, just load a function from this file or
            # one that has be explicitly imported.
            function = getattr( globals(), filter_name.strip() )
        return function

    def _import_filter( self, module_name, function_name ):
        function_name = function_name.strip()
        for base_module in self.__base_modules:
            full_module_name = "%s.%s" % ( base_module, module_name.strip() )
            try:
                __import__( full_module_name )
            except ImportError:
                # log.debug("Failed to load module.", exc_info=True)
                continue
            module = sys.modules[ full_module_name ]
            if hasattr( module, function_name ):
                return getattr( module, function_name )
        raise Exception("Failed to find filter %s.%s" % (module_name, function_name))


# Stock Filter Functions
def _not_hidden( context, tool ):
    return not tool.hidden


def _handle_authorization( context, tool ):
    user = context.trans.user
    if tool.require_login and not user:
        return False
    if not tool.allow_user_access( user, attempting_access=False ):
        return False
    return True


def _has_trackster_conf( context, tool ):
    return tool.trackster_conf



"""
Classes for wrapping Objects and Sanitizing string output.
"""
import collections
import inspect
import logging
import string
import sys
from numbers import Number

try:
    from types import NoneType
except ImportError:
    NoneType = type(None)
try:
    from types import NotImplementedType
except ImportError:
    NotImplementedType = type(NotImplemented)

try:
    from types import EllipsisType
except ImportError:
    EllipsisType = type(Ellipsis)

try:
    from types import XRangeType
except ImportError:
    XRangeType = range

try:
    from types import SliceType
except ImportError:
    SliceType = slice

try:
    from types import (
        BufferType,
        DictProxyType
    )
except ImportError:
    # Py3 doesn't have these concepts, just treat them like SliceType that
    # so they are __WRAP_NO_SUBCLASS__.
    BufferType = SliceType
    DictProxyType = SliceType

from types import (
    BuiltinFunctionType,
    BuiltinMethodType,
    CodeType,
    FrameType,
    FunctionType,
    GeneratorType,
    GetSetDescriptorType,
    MemberDescriptorType,
    MethodType,
    ModuleType,
    TracebackType,
)

from six.moves import (
    copyreg as copy_reg,
    UserDict
)

from galaxy.util import sanitize_lists_to_string as _sanitize_lists_to_string

log = logging.getLogger( __name__ )

# Define different behaviors for different types, see also: https://docs.python.org/2/library/types.html

# Known Callable types
__CALLABLE_TYPES__ = ( FunctionType, MethodType, GeneratorType, CodeType, BuiltinFunctionType, BuiltinMethodType, )

# Always wrap these types without attempting to subclass
__WRAP_NO_SUBCLASS__ = ( ModuleType, XRangeType, SliceType, BufferType, TracebackType, FrameType, DictProxyType,
                         GetSetDescriptorType, MemberDescriptorType ) + __CALLABLE_TYPES__

# Don't wrap or sanitize.
__DONT_SANITIZE_TYPES__ = ( Number, bool, NoneType, NotImplementedType, EllipsisType, bytearray, )

# Don't wrap, but do sanitize.
__DONT_WRAP_TYPES__ = tuple()  # ( basestring, ) so that we can get the unsanitized string, we will now wrap basestring instances

# Wrap contents, but not the container
__WRAP_SEQUENCES__ = ( tuple, list, )
__WRAP_SETS__ = ( set, frozenset, )
__WRAP_MAPPINGS__ = ( dict, UserDict, )


# Define the set of characters that are not sanitized, and define a set of mappings for those that are.
# characters that are valid
VALID_CHARACTERS = set( string.ascii_letters + string.digits + " -=_.()/+*^,:?!@" )

# characters that are allowed but need to be escaped
CHARACTER_MAP = { '>': '__gt__',
                  '<': '__lt__',
                  "'": '__sq__',
                  '"': '__dq__',
                  '[': '__ob__',
                  ']': '__cb__',
                  '{': '__oc__',
                  '}': '__cc__',
                  '\n': '__cn__',
                  '\r': '__cr__',
                  '\t': '__tc__',
                  '#': '__pd__'}

INVALID_CHARACTER = "X"

if sys.version_info > (3, 0):
    # __coerce__ doesn't do anything under Python anyway.
    def coerce(x, y):
        return x


def cmp(x, y):
    # Builtin in Python 2, but not Python 3.
    return (x > y) - (x < y)


def sanitize_lists_to_string( values, valid_characters=VALID_CHARACTERS, character_map=CHARACTER_MAP, invalid_character=INVALID_CHARACTER  ):
    return _sanitize_lists_to_string( values, valid_characters=valid_characters, character_map=character_map, invalid_character=invalid_character  )


def wrap_with_safe_string( value, no_wrap_classes=None ):
    """
    Recursively wrap values that should be wrapped.
    """

    def __do_wrap( value ):
        if isinstance( value, SafeStringWrapper ):
            # Only ever wrap one-layer
            return value
        if isinstance( value, collections.Callable ):
            safe_class = CallableSafeStringWrapper
        else:
            safe_class = SafeStringWrapper
        if isinstance( value, no_wrap_classes ):
            return value
        if isinstance( value, __DONT_WRAP_TYPES__ ):
            return sanitize_lists_to_string( value, valid_characters=VALID_CHARACTERS, character_map=CHARACTER_MAP )
        if isinstance( value, __WRAP_NO_SUBCLASS__ ):
            return safe_class( value, safe_string_wrapper_function=__do_wrap )
        for this_type in __WRAP_SEQUENCES__ + __WRAP_SETS__:
            if isinstance( value, this_type ):
                return this_type( map( __do_wrap, value ) )
        for this_type in __WRAP_MAPPINGS__:
            if isinstance( value, this_type ):
                # Wrap both key and value
                return this_type( ( __do_wrap( x[0] ), __do_wrap( x[1] ) ) for x in value.items() )
        # Create a dynamic class that joins SafeStringWrapper with the object being wrapped.
        # This allows e.g. isinstance to continue to work.
        try:
            wrapped_class_name = value.__name__
            wrapped_class = value
        except:
            wrapped_class_name = value.__class__.__name__
            wrapped_class = value.__class__
        value_mod = inspect.getmodule( value )
        if value_mod:
            wrapped_class_name = "%s.%s" % ( value_mod.__name__, wrapped_class_name )
        wrapped_class_name = "SafeStringWrapper(%s:%s)" % ( wrapped_class_name, ",".join( sorted( map( str, no_wrap_classes ) ) ) )
        do_wrap_func_name = "__do_wrap_%s" % ( wrapped_class_name )
        do_wrap_func = __do_wrap
        global_dict = globals()
        if wrapped_class_name in global_dict:
            # Check to see if we have created a wrapper for this class yet, if so, reuse
            wrapped_class = global_dict.get( wrapped_class_name )
            do_wrap_func = global_dict.get( do_wrap_func_name, __do_wrap )
        else:
            try:
                wrapped_class = type( wrapped_class_name, ( safe_class, wrapped_class, ), {} )
            except TypeError as e:
                # Fail-safe for when a class cannot be dynamically subclassed.
                log.warning( "Unable to create dynamic subclass for %s, %s: %s", type( value), value, e )
                wrapped_class = type( wrapped_class_name, ( safe_class, ), {} )
            if wrapped_class not in ( SafeStringWrapper, CallableSafeStringWrapper ):
                # Save this wrapper for reuse and pickling/copying
                global_dict[ wrapped_class_name ] = wrapped_class
                do_wrap_func.__name__ = do_wrap_func_name
                global_dict[ do_wrap_func_name ] = do_wrap_func

                def pickle_safe_object( safe_object ):
                    return ( wrapped_class, ( safe_object.unsanitized, do_wrap_func, ) )
                # Set pickle and copy properties
                copy_reg.pickle( wrapped_class, pickle_safe_object, do_wrap_func )
        return wrapped_class( value, safe_string_wrapper_function=do_wrap_func )

    # Determine classes not to wrap
    if no_wrap_classes:
        if not isinstance( no_wrap_classes, ( tuple, list ) ):
            no_wrap_classes = [ no_wrap_classes ]
        no_wrap_classes = list( no_wrap_classes ) + list( __DONT_SANITIZE_TYPES__ ) + [ SafeStringWrapper ]
    else:
        no_wrap_classes = list( __DONT_SANITIZE_TYPES__ ) + [ SafeStringWrapper ]
    no_wrap_classes = tuple( set( sorted( no_wrap_classes, key=str ) ) )
    return __do_wrap( value )


# N.B. refer to e.g. https://docs.python.org/2/reference/datamodel.html for information on Python's Data Model.


class SafeStringWrapper( object ):
    """
    Class that wraps and sanitizes any provided value's attributes
    that will attempt to be cast into a string.

    Attempts to mimic behavior of original class, including operands.

    To ensure proper handling of e.g. subclass checks, the *wrap_with_safe_string()*
    method should be used.

    This wrapping occurs in a recursive/parasitic fashion, as all called attributes of
    the originally wrapped object will also be wrapped and sanitized, unless the attribute
    is of a type found in __DONT_SANITIZE_TYPES__ + __DONT_WRAP_TYPES__, where e.g. ~(strings
    will still be sanitized, but not wrapped), and e.g. integers will have neither.
    """
    __UNSANITIZED_ATTRIBUTE_NAME__ = 'unsanitized'
    __NO_WRAP_NAMES__ = [ '__safe_string_wrapper_function__', '__class__', __UNSANITIZED_ATTRIBUTE_NAME__]

    def __new__( cls, *arg, **kwd ):
        # We need to define a __new__ since, we are subclassing from e.g. immutable str, which internally sets data
        # that will be used when other + this (this + other is handled by __add__)
        try:
            return super( SafeStringWrapper, cls ).__new__( cls, sanitize_lists_to_string( arg[0], valid_characters=VALID_CHARACTERS, character_map=CHARACTER_MAP ) )
        except Exception as e:
            log.warning( "Could not provide an argument to %s.__new__: %s; will try without arguments.", cls, e )
            return super( SafeStringWrapper, cls ).__new__( cls )

    def __init__( self, value, safe_string_wrapper_function=wrap_with_safe_string ):
        self.unsanitized = value
        self.__safe_string_wrapper_function__ = safe_string_wrapper_function

    def __str__( self ):
        return sanitize_lists_to_string( self.unsanitized, valid_characters=VALID_CHARACTERS, character_map=CHARACTER_MAP )

    def __repr__( self ):
        return "%s object at %x on: %s" % ( sanitize_lists_to_string( self.__class__.__name__, valid_characters=VALID_CHARACTERS, character_map=CHARACTER_MAP ), id( self ), sanitize_lists_to_string( repr( self.unsanitized ), valid_characters=VALID_CHARACTERS, character_map=CHARACTER_MAP ) )

    def __lt__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.unsanitized < other

    def __le__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.unsanitized <= other

    def __eq__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.unsanitized == other

    def __ne__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.unsanitized != other

    def __gt__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.unsanitized > other

    def __ge__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.unsanitized >= other

    def __cmp__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return cmp( self.unsanitized, other )

    # Do not implement __rcmp__, python 2.2 < 2.6

    def __hash__( self ):
        return hash( self.unsanitized )

    def __bool__( self ):
        return bool( self.unsanitized )
    __nonzero__ = __bool__

    # Do not implement __unicode__, we will rely on __str__

    def __getattr__( self, name ):
        if name in SafeStringWrapper.__NO_WRAP_NAMES__:
            # FIXME: is this ever reached?
            return object.__getattribute__( self, name )
        return self.__safe_string_wrapper_function__( getattr( self.unsanitized, name ) )

    def __setattr__( self, name, value ):
        if name in SafeStringWrapper.__NO_WRAP_NAMES__:
            return object.__setattr__( self, name, value )
        return setattr( self.unsanitized, name, value )

    def __delattr__( self, name ):
        if name in SafeStringWrapper.__NO_WRAP_NAMES__:
            return object.__delattr__( self, name )
        return delattr( self.unsanitized, name )

    def __getattribute__( self, name ):
        if name in SafeStringWrapper.__NO_WRAP_NAMES__:
            return object.__getattribute__( self, name )
        return self.__safe_string_wrapper_function__( getattr( object.__getattribute__( self, 'unsanitized' ), name ) )

    # Skip Descriptors

    # Skip __slots__

    # Don't need to define a metaclass, we'll use the helper function to handle with subclassing for e.g. isinstance()

    # Revisit:
    # __instancecheck__
    # __subclasscheck__
    # We are using a helper class to create dynamic subclasses to handle class checks

    # We address __call__ as needed based upon unsanitized, through the use of a CallableSafeStringWrapper class

    def __len__( self ):
        original_value = self.unsanitized
        while isinstance( original_value, SafeStringWrapper ):
            original_value = self.unsanitized
        return len( self.unsanitized )

    def __getitem__( self, key ):
        return self.__safe_string_wrapper_function__( self.unsanitized[ key ] )

    def __setitem__( self, key, value ):
        while isinstance( value, SafeStringWrapper ):
            value = value.unsanitized
        self.unsanitized[ key ] = value

    def __delitem__( self, key ):
        del self.unsanitized[ key ]

    def __iter__( self ):
        return iter( map( self.__safe_string_wrapper_function__, iter( self.unsanitized ) ) )

    # Do not implement __reversed__

    def __contains__( self, item ):
        # FIXME: Do we need to consider if item is/isn't or does/doesn't contain SafeStringWrapper?
        # When considering e.g. nested lists/dicts/etc, this gets complicated
        while isinstance( item, SafeStringWrapper ):
            item = item.unsanitized
        return item in self.unsanitized

    # Not sure that we need these slice methods, but will provide anyway
    def __getslice__( self, i, j ):
        return self.__safe_string_wrapper_function__( self.unsanitized[ i:j ] )

    def __setslice__( self, i, j, value ):
        self.unsanitized[ i:j ] = value

    def __delslice__( self, i, j ):
        del self.unsanitized[ i:j ]

    def __add__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( self.unsanitized + other )

    def __sub__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( self.unsanitized - other )

    def __mul__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( self.unsanitized * other )

    def __floordiv__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( self.unsanitized // other )

    def __mod__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( self.unsanitized % other )

    def __divmod__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( divmod( self.unsanitized, other ) )

    def __pow__( self, *other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( pow( self.unsanitized, *other ) )

    def __lshift__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( self.unsanitized << other )

    def __rshift__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( self.unsanitized >> other )

    def __and__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( self.unsanitized & other )

    def __xor__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( self.unsanitized ^ other )

    def __or__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( self.unsanitized | other )

    def __div__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( self.unsanitized / other )

    def __truediv__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( self.unsanitized / other )

    # The only reflected operand that we will define is __rpow__, due to coercion rules complications as per docs
    def __rpow__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return self.__safe_string_wrapper_function__( pow( other, self.unsanitized ) )

    # Do not implement in-place operands

    def __neg__( self ):
        return self.__safe_string_wrapper_function__( -self.unsanitized )

    def __pos__( self ):
        return self.__safe_string_wrapper_function__( +self.unsanitized )

    def __abs__( self ):
        return self.__safe_string_wrapper_function__( abs( self.unsanitized ) )

    def __invert__( self ):
        return self.__safe_string_wrapper_function__( ~self.unsanitized )

    def __complex__( self ):
        return self.__safe_string_wrapper_function__( complex( self.unsanitized ) )

    def __int__( self ):
        return int( self.unsanitized )

    def __float__( self ):
        return float( self.unsanitized )

    def __oct__( self ):
        return oct( self.unsanitized )

    def __hex__( self ):
        return hex( self.unsanitized )

    def __index__( self ):
        return self.unsanitized.index()

    def __coerce__( self, other ):
        while isinstance( other, SafeStringWrapper ):
            other = other.unsanitized
        return coerce( self.unsanitized, other )

    def __enter__( self ):
        return self.unsanitized.__enter__()

    def __exit__( self, *args ):
        return self.unsanitized.__exit__( *args )


class CallableSafeStringWrapper( SafeStringWrapper ):

    def __call__( self, *args, **kwds ):
        return self.__safe_string_wrapper_function__( self.unsanitized( *args, **kwds ) )


# Enable pickling/deepcopy
def pickle_SafeStringWrapper( safe_object ):
    args = ( safe_object.unsanitized, )
    cls = SafeStringWrapper
    if isinstance( safe_object, CallableSafeStringWrapper ):
        cls = CallableSafeStringWrapper
    return ( cls, args )


copy_reg.pickle( SafeStringWrapper, pickle_SafeStringWrapper, wrap_with_safe_string )
copy_reg.pickle( CallableSafeStringWrapper, pickle_SafeStringWrapper, wrap_with_safe_string )



from xml.etree import ElementTree

try:
    import yaml
except ImportError:
    yaml = None

from galaxy.util.submodules import submodules


def plugins_dict(module, plugin_type_identifier):
    """ Walk through all classes in submodules of module and find ones labelled
    with specified plugin_type_identifier and throw in a dictionary to allow
    constructions from plugins by these types later on.
    """
    plugin_dict = {}

    for plugin_module in submodules( module ):
        # FIXME: this is not how one is suppose to use __all__ why did you do
        # this past John?
        for clazz in getattr( plugin_module, "__all__", [] ):
            try:
                clazz = getattr( plugin_module, clazz )
            except TypeError:
                clazz = clazz
            plugin_type = getattr( clazz, plugin_type_identifier, None )
            if plugin_type:
                plugin_dict[ plugin_type ] = clazz

    return plugin_dict


def load_plugins(plugins_dict, plugin_source, extra_kwds={}):
    source_type, source = plugin_source
    if source_type == "xml":
        return __load_plugins_from_element(plugins_dict, source, extra_kwds)
    else:
        return __load_plugins_from_dicts(plugins_dict, source, extra_kwds)


def __load_plugins_from_element(plugins_dict, plugins_element, extra_kwds):
    plugins = []

    for plugin_element in plugins_element.getchildren():
        plugin_type = plugin_element.tag
        plugin_kwds = dict( plugin_element.items() )
        plugin_kwds.update( extra_kwds )
        plugin = plugins_dict[ plugin_type ]( **plugin_kwds )
        plugins.append( plugin )

    return plugins


def __load_plugins_from_dicts(plugins_dict, configs, extra_kwds):
    plugins = []

    for config in configs:
        plugin_type = config[ "type" ]
        plugin_kwds = config
        plugin_kwds.update( extra_kwds )
        plugin = plugins_dict[ plugin_type ]( **plugin_kwds )
        plugins.append( plugin )

    return plugins


def plugin_source_from_path(path):
    if path.endswith(".yaml") or path.endswith(".yml"):
        return ('dict', __read_yaml(path))
    else:
        return ('xml', ElementTree.parse( path ).getroot())


def __read_yaml(path):
    if yaml is None:
        raise ImportError("Attempting to read YAML configuration file - but PyYAML dependency unavailable.")

    with open(path, "rb") as f:
        return yaml.load(f)



""" Handle details of tool tagging - perhaps a deprecated feature.
"""
import logging
from abc import (
    ABCMeta,
    abstractmethod
)

import six

log = logging.getLogger( __name__ )


def tool_tag_manager( app ):
    """ Build a tool tag manager according to app's configuration
    and return it.
    """
    if hasattr( app.config, "get_bool" ) and app.config.get_bool( 'enable_tool_tags', False ):
        return PersistentToolTagManager( app )
    else:
        return NullToolTagManager()


@six.add_metaclass(ABCMeta)
class AbstractToolTagManager( object ):

    @abstractmethod
    def reset_tags( self ):
        """ Starting to load tool panels, reset all tags.
        """

    @abstractmethod
    def handle_tags( self, tool_id, tool_definition_source ):
        """ Parse out tags and persist them.
        """


class NullToolTagManager( AbstractToolTagManager ):

    def reset_tags( self ):
        return None

    def handle_tags( self, tool_id, tool_definition_source ):
        return None


class PersistentToolTagManager( AbstractToolTagManager ):

    def __init__( self, app ):
        self.app = app
        self.sa_session = app.model.context

    def reset_tags( self ):
        log.info("removing all tool tag associations (" + str( self.sa_session.query( self.app.model.ToolTagAssociation ).count() ) + ")" )
        self.sa_session.query( self.app.model.ToolTagAssociation ).delete()
        self.sa_session.flush()

    def handle_tags( self, tool_id, tool_definition_source ):
        elem = tool_definition_source
        if self.app.config.get_bool( 'enable_tool_tags', False ):
            tag_names = elem.get( "tags", "" ).split( "," )
            for tag_name in tag_names:
                if tag_name == '':
                    continue
                tag = self.sa_session.query( self.app.model.Tag ).filter_by( name=tag_name ).first()
                if not tag:
                    tag = self.app.model.Tag( name=tag_name )
                    self.sa_session.add( tag )
                    self.sa_session.flush()
                    tta = self.app.model.ToolTagAssociation( tool_id=tool_id, tag_id=tag.id )
                    self.sa_session.add( tta )
                    self.sa_session.flush()
                else:
                    for tagged_tool in tag.tagged_tools:
                        if tagged_tool.tool_id == tool_id:
                            break
                    else:
                        tta = self.app.model.ToolTagAssociation( tool_id=tool_id, tag_id=tag.id )
                        self.sa_session.add( tta )
                        self.sa_session.flush()



try:
    import requests
except ImportError:
    requests = None
import yaml

from six.moves.urllib.parse import quote

from ..locations import (
    ToolLocationResolver,
)


class DockStoreResolver(ToolLocationResolver):

    scheme = "dockstore"

    def get_tool_source_path(self, uri_like):
        assert uri_like.startswith("dockstore://")
        tool_id = uri_like[len("dockstore://"):]
        if ":" in tool_id:
            tool_id, version = tool_id.split(":", 1)
        else:
            tool_id, version = tool_id, "latest"
        tmp_path = self._temp_path(uri_like + ".cwl")
        cwl_str = _Ga4ghToolClient().get_tool_cwl(tool_id, version=version, as_string=True)
        with open(tmp_path, "wb") as f:
            f.write(cwl_str)
        return tmp_path


class _Ga4ghToolClient(object):

    def __init__(self, base_url="https://www.dockstore.org:8443/api"):
        self.base_url = base_url

    def get_tools(self):
        return self._requests.get("%s/ga4gh/v1/tools" % self.base_url)

    def get_tool(self, tool_id):
        url = "%s/ga4gh/v1/tools/%s" % (self.base_url, quote(tool_id, safe=''))
        return self._requests.get(url)

    def get_tool_version(self, tool_id, version="latest"):
        url = "%s/ga4gh/v1/tools/%s/versions/%s" % (self.base_url, quote(tool_id, safe=''), version)
        return self._requests.get(url)

    def get_tool_descriptor(self, tool_id, version="latest", tool_type="CWL"):
        url = "%s/ga4gh/v1/tools/%s/versions/%s/%s/descriptor" % (self.base_url, quote(tool_id, safe=''), version, tool_type)
        return self._requests.get(url)

    def get_tool_cwl(self, tool_id, version="latest", as_string=False):
        tool_type = "CWL"
        url = "%s/ga4gh/v1/tools/%s/versions/%s/%s/descriptor" % (self.base_url, quote(tool_id, safe=''), version, tool_type)
        descriptor_response = self._requests.get(url)
        descriptor_str = descriptor_response.json()["descriptor"]
        if as_string:
            return descriptor_str
        else:
            return yaml.load(descriptor_str)

    @property
    def _requests(self):
        if requests is None:
            raise Exception("requests Python library needs to be installed use GA4GH APIs")
        return requests


__all__ = ("DockStoreResolver",)



"""This module describes the :class:`ExplicitContainerResolver` ContainerResolver plugin."""
import logging

from ..container_resolvers import (
    ContainerResolver,
)

log = logging.getLogger(__name__)


class ExplicitContainerResolver(ContainerResolver):
    """Find explicit containers referenced in the tool description (e.g. tool XML file) if present."""

    resolver_type = "explicit"

    def resolve(self, enabled_container_types, tool_info):
        """Find a container explicitly mentioned in tool description.

        This ignores the tool requirements and assumes the tool author crafted
        a correct container.
        """
        for container_description in tool_info.container_descriptions:
            if self._container_type_enabled(container_description, enabled_container_types):
                return container_description

        return None



""" Modules will run collectl in playback mode and collect various process
statistics for a given pid's process and process ancestors.
"""
import collections
import csv
import logging
import sys
import tempfile

from galaxy import util

from ..collectl import stats

if sys.version_info > (3,):
    long = int

log = logging.getLogger( __name__ )

# Collectl process information cheat sheet:
#
# Record process information for current user.
# %  collectl -sZ -f./__instrument_collectl  -i 10:10 --procfilt U$USER
#
# TSV Replay of processing information in plottable mode...
#
# % collectl -sZ -P --sep=9 -p __instrument_collectl-jlaptop13-20140322-120919.raw.gz
#
# Has following columns:
#   Date   Time    PID     User    PR      PPID    THRD    S       VmSize  VmLck   VmRSS   VmData  VmStk   VmExe   VmLib   CPU       SysT    UsrT    PCT     AccumT  RKB     WKB     RKBC    WKBC    RSYS    WSYS    CNCL    MajF    MinF    Command
#

# Process data dumped one row per process per interval.
# http://collectl.sourceforge.net/Data-detail.html
PROCESS_COLUMNS = [
    "#Date",  # Date of interval - e.g. 20140322
    "Time",  # Time of interval - 12:18:58
    "PID",  # Process pid.
    "User",  # Process user.
    "PR",  # Priority of process.
    "PPID",  # Parent PID of process.
    "THRD",  # Thread???
    "S",  # Process state - S - Sleeping, D - Uninterruptable Sleep, R - Running, Z - Zombie or T - Stopped/Traced
    # Memory options - http://ewx.livejournal.com/579283.html
    "VmSize",
    "VmLck",
    "VmRSS",
    "VmData",
    "VmStk",
    "VmExe",
    "VmLib",
    "CPU",  # CPU number of process
    "SysT",  # Amount of system time consumed during interval
    "UsrT",  # Amount user time consumed during interval
    "PCT",  # Percentage of current interval consumed by task
    "AccumT",  # Total accumulated System and User time since the process began execution
    # kilobytes read/written - requires I/O level monitoring to be enabled in kernel.
    "RKB",  # kilobytes read by process - requires I/O monitoring in kernel
    "WKB",
    "RKBC",
    "WKBC",
    "RSYS",  # Number of read system calls
    "WSYS",  # Number of write system calls
    "CNCL",
    "MajF",  # Number of major page faults
    "MinF",  # Number of minor page faults
    "Command",  # Command executed
]

# Types of statistics this module can summarize
STATISTIC_TYPES = [ "max", "min", "sum", "count", "avg" ]

COLUMN_INDICES = dict( [ ( col, i ) for i, col in enumerate( PROCESS_COLUMNS ) ] )
PID_INDEX = COLUMN_INDICES[ "PID" ]
PARENT_PID_INDEX = COLUMN_INDICES[ "PPID" ]

DEFAULT_STATISTICS = [
    ("max", "VmSize"),
    ("avg", "VmSize"),
    ("max", "VmRSS"),
    ("avg", "VmRSS"),
    ("sum", "SysT"),
    ("sum", "UsrT"),
    ("max", "PCT"),
    ("avg", "PCT"),
    ("max", "AccumT"),
    ("sum", "RSYS"),
    ("sum", "WSYS"),
]


def parse_process_statistics( statistics ):
    """ Turn string or list of strings into list of tuples in format ( stat,
    resource ) where stat is a value from STATISTIC_TYPES and resource is a
    value from PROCESS_COLUMNS.
    """
    if statistics is None:
        statistics = DEFAULT_STATISTICS

    statistics = util.listify( statistics )
    statistics = [ _tuplize_statistic(_) for _ in statistics ]
    # Check for validity...
    for statistic in statistics:
        if statistic[ 0 ] not in STATISTIC_TYPES:
            raise Exception( "Unknown statistic type encountered %s" % statistic[ 0 ] )
        if statistic[ 1 ] not in PROCESS_COLUMNS:
            raise Exception( "Unknown process column encountered %s" % statistic[ 1 ] )
    return statistics


def generate_process_statistics( collectl_playback_cli, pid, statistics=DEFAULT_STATISTICS ):
    """ Playback collectl file and generate summary statistics.
    """
    with tempfile.NamedTemporaryFile( ) as tmp_tsv:
        collectl_playback_cli.run( stdout=tmp_tsv )
        with open( tmp_tsv.name, "r" ) as tsv_file:
            return _read_process_statistics( tsv_file, pid, statistics )


def _read_process_statistics( tsv_file, pid, statistics ):
    process_summarizer = CollectlProcessSummarizer( pid, statistics )
    current_interval = None

    for row in csv.reader( tsv_file, dialect="excel-tab" ):
        if current_interval is None:
            for header, expected_header in zip( row, PROCESS_COLUMNS ):
                if header.lower() != expected_header.lower():
                    raise Exception( "Unknown header value encountered while processing collectl playback - %s" % header )

            # First row, check contains correct header.
            current_interval = CollectlProcessInterval()
            continue

        if current_interval.row_is_in( row ):
            current_interval.add_row( row )
        else:
            process_summarizer.handle_interval( current_interval )
            current_interval = CollectlProcessInterval()

    # Do we have unsummarized rows...
    if current_interval and current_interval.rows:
        process_summarizer.handle_interval( current_interval )

    return process_summarizer.get_statistics()


class CollectlProcessSummarizer( object ):

    def __init__( self, pid, statistics ):
        self.pid = pid
        self.statistics = statistics
        self.columns_of_interest = set( [ s[ 1 ] for s in statistics ] )
        self.tree_statistics = collections.defaultdict( stats.StatisticsTracker )
        self.process_accum_statistics = collections.defaultdict( stats.StatisticsTracker )
        self.interval_count = 0

    def handle_interval( self, interval ):
        self.interval_count += 1
        rows = self.__rows_for_process( interval.rows, self.pid )
        for column_name in self.columns_of_interest:
            column_index = COLUMN_INDICES[ column_name ]

            if column_name == "AccumT":
                # Should not sum this across pids each interval, sum max at end...
                for r in rows:
                    pid_seconds = self.__time_to_seconds( r[ column_index ] )
                    self.process_accum_statistics[ r[ PID_INDEX ] ].track( pid_seconds )
            else:
                # All other stastics should be summed across whole process tree
                # at each interval I guess.
                if column_name in [ "SysT", "UsrT", "PCT" ]:
                    to_num = float
                else:
                    to_num = long

                interval_stat = sum( to_num( r[ column_index ] ) for r in rows )
                self.tree_statistics[ column_name ].track( interval_stat )

    def get_statistics( self ):
        if self.interval_count == 0:
            return []

        computed_statistics = []
        for statistic in self.statistics:
            statistic_type, column = statistic
            if column == "AccumT":
                # Only thing that makes sense is sum
                if statistic_type != "max":
                    log.warning( "Only statistic max makes sense for AccumT" )
                    continue

                value = sum( v.max for v in self.process_accum_statistics.values() )
            else:
                statistics_tracker = self.tree_statistics[ column ]
                value = getattr( statistics_tracker, statistic_type )

            computed_statistic = ( statistic, value )
            computed_statistics.append( computed_statistic )

        return computed_statistics

    def __rows_for_process( self, rows, pid ):
        process_rows = []
        pids = self.__all_child_pids( rows, pid )
        for row in rows:
            if row[ PID_INDEX ] in pids:
                process_rows.append( row )
        return process_rows

    def __all_child_pids( self, rows, pid ):
        pids_in_process_tree = set( [ str( self.pid ) ] )
        added = True
        while added:
            added = False
            for row in rows:
                pid = row[ PID_INDEX ]
                parent_pid = row[ PARENT_PID_INDEX ]
                if parent_pid in pids_in_process_tree and pid not in pids_in_process_tree:
                    pids_in_process_tree.add( pid )
                    added = True
        return pids_in_process_tree

    def __time_to_seconds( self, minutes_str ):
        parts = minutes_str.split( ":" )
        seconds = 0.0
        for i, val in enumerate( parts ):
            seconds += float(val) * ( 60 ** ( len( parts ) - ( i + 1 ) ) )
        return seconds


class CollectlProcessInterval( object ):
    """ Represent all rows in collectl playback file for given time slice with
    ability to filter out just rows corresponding to the process tree
    corresponding to a given pid.
    """

    def __init__( self ):
        self.rows = []

    def row_is_in( self, row ):
        if not self.rows:  # No rows, this row defines interval.
            return True
        first_row = self.rows[ 0 ]
        return first_row[ 0 ] == row[ 0 ] and first_row[ 1 ] == row[ 1 ]

    def add_row( self, row ):
        self.rows.append( row )


def _tuplize_statistic( statistic ):
    if not isinstance( statistic, tuple ):
        statistic_split = statistic.split( "_", 1 )
        statistic = ( statistic_split[ 0 ].lower(), statistic_split[ 1 ] )
    return statistic


__all__ = ( 'generate_process_statistics', )



import os
import sys
import threading
import time
import traceback

from six import iteritems


def get_current_thread_object_dict():
    """
    Get a dictionary of all 'Thread' objects created via the threading
    module keyed by thread_id. Note that not all interpreter threads
    have a thread objects, only the main thread and any created via the
    'threading' module. Threads created via the low level 'thread' module
    will not be in the returned dictionary.

    HACK: This mucks with the internals of the threading module since that
          module does not expose any way to match 'Thread' objects with
          intepreter thread identifiers (though it should).
    """
    rval = dict()
    # Acquire the lock and then union the contents of 'active' and 'limbo'
    # threads into the return value.
    threading._active_limbo_lock.acquire()
    rval.update( threading._active )
    rval.update( threading._limbo )
    threading._active_limbo_lock.release()
    return rval


class Heartbeat( threading.Thread ):
    """
    Thread that periodically dumps the state of all threads to a file
    """
    def __init__( self, config, name="Heartbeat Thread", period=20, fname="heartbeat.log" ):
        threading.Thread.__init__( self, name=name )
        self.config = config
        self.should_stop = False
        self.period = period
        self.fname = fname
        self.file = None
        self.fname_nonsleeping = None
        self.file_nonsleeping = None
        self.pid = None
        self.nonsleeping_heartbeats = { }
        # Event to wait on when sleeping, allows us to interrupt for shutdown
        self.wait_event = threading.Event()

    def run( self ):
        self.pid = os.getpid()
        self.fname = self.fname.format(
            server_name=self.config.server_name,
            pid=self.pid
        )
        fname, ext = os.path.splitext(self.fname)
        self.fname_nonsleeping = fname + '.nonsleeping' + ext
        wait = self.period
        if self.period <= 0:
            wait = 60
        while not self.should_stop:
            if self.period > 0:
                self.dump()
            self.wait_event.wait( wait )

    def open_logs( self ):
        if self.file is None or self.file.closed:
            self.file = open( self.fname, "a" )
            self.file_nonsleeping = open( self.fname_nonsleeping, "a" )
            self.file.write( "Heartbeat for pid %d thread started at %s\n\n" % ( self.pid, time.asctime() ) )
            self.file_nonsleeping.write( "Non-Sleeping-threads for pid %d thread started at %s\n\n" % ( self.pid, time.asctime() ) )

    def close_logs( self ):
        if self.file is not None and not self.file.closed:
            self.file.write( "Heartbeat for pid %d thread stopped at %s\n\n" % ( self.pid, time.asctime() ) )
            self.file_nonsleeping.write( "Non-Sleeping-threads for pid %d thread stopped at %s\n\n" % ( self.pid, time.asctime() ) )
            self.file.close()
            self.file_nonsleeping.close()

    def dump( self ):
        self.open_logs()
        try:
            # Print separator with timestamp
            self.file.write( "Traceback dump for all threads at %s:\n\n" % time.asctime() )
            # Print the thread states
            threads = get_current_thread_object_dict()
            for thread_id, frame in iteritems(sys._current_frames()):
                if thread_id in threads:
                    object = repr( threads[thread_id] )
                else:
                    object = "<No Thread object>"
                self.file.write( "Thread %s, %s:\n\n" % ( thread_id, object ) )
                traceback.print_stack( frame, file=self.file )
                self.file.write( "\n" )
            self.file.write( "End dump\n\n" )
            self.file.flush()
            self.print_nonsleeping(threads)
        except:
            self.file.write( "Caught exception attempting to dump thread states:" )
            traceback.print_exc( None, self.file )
            self.file.write( "\n" )

    def shutdown( self ):
        self.should_stop = True
        self.wait_event.set()
        self.close_logs()
        self.join()

    def thread_is_sleeping( self, last_stack_frame ):
        """
        Returns True if the given stack-frame represents a known
        sleeper function (at least in python 2.5)
        """
        _filename = last_stack_frame[0]
        # _line = last_stack_frame[1]
        _funcname = last_stack_frame[2]
        _text = last_stack_frame[3]
        # Ugly hack to tell if a thread is supposedly sleeping or not
        # These are the most common sleeping functions I've found.
        # Is there a better way? (python interpreter internals?)
        # Tested only with python 2.5
        if _funcname == "wait" and _text == "waiter.acquire()":
            return True
        if _funcname == "wait" and _text == "_sleep(delay)":
            return True
        if _funcname == "accept" and _text[-14:] == "_sock.accept()":
            return True
        if _funcname in ("monitor", "__monitor", "app_loop", "check") \
                and _text.startswith("time.sleep(") and _text.endswith(")"):
            return True
        if _funcname == "drain_events" and _text == "sleep(polling_interval)":
            return True
        # Ugly hack: always skip the heartbeat thread
        # TODO: get the current thread-id in python
        #   skip heartbeat thread by thread-id, not by filename
        if _filename.find("/lib/galaxy/util/heartbeat.py") != -1:
            return True
        # By default, assume the thread is not sleeping
        return False

    def get_interesting_stack_frame( self, stack_frames ):
        """
        Scans a given backtrace stack frames, returns a single
        quadraple of [filename, line, function-name, text] of
        the single, deepest, most interesting frame.

        Interesting being::

          inside the galaxy source code ("/lib/galaxy"),
          prefreably not an egg.
        """
        for _filename, _line, _funcname, _text in reversed(stack_frames):
            idx = _filename.find("/lib/galaxy/")
            if idx != -1:
                relative_filename = _filename[idx:]
                return ( relative_filename, _line, _funcname, _text )
        # no "/lib/galaxy" code found, return the innermost frame
        return stack_frames[-1]

    def print_nonsleeping( self, threads_object_dict ):
        self.file_nonsleeping.write( "Non-Sleeping threads at %s:\n\n" % time.asctime() )
        all_threads_are_sleeping = True
        threads = get_current_thread_object_dict()
        for thread_id, frame in iteritems(sys._current_frames()):
            if thread_id in threads:
                object = repr( threads[thread_id] )
            else:
                object = "<No Thread object>"
            tb = traceback.extract_stack(frame)
            if self.thread_is_sleeping(tb[-1]):
                if thread_id in self.nonsleeping_heartbeats:
                    del self.nonsleeping_heartbeats[thread_id]
                continue

            # Count non-sleeping thread heartbeats
            if thread_id in self.nonsleeping_heartbeats:
                self.nonsleeping_heartbeats[thread_id] += 1
            else:
                self.nonsleeping_heartbeats[thread_id] = 1

            good_frame = self.get_interesting_stack_frame(tb)
            self.file_nonsleeping.write( "Thread %s\t%s\tnon-sleeping for %d heartbeat(s)\n  File %s:%d\n    Function \"%s\"\n      %s\n" %
                ( thread_id, object, self.nonsleeping_heartbeats[thread_id], good_frame[0], good_frame[1], good_frame[2], good_frame[3] ) )
            all_threads_are_sleeping = False

        if all_threads_are_sleeping:
            self.file_nonsleeping.write( "All threads are sleeping.\n" )
        self.file_nonsleeping.write( "\n" )
        self.file_nonsleeping.flush()

    def dump_signal_handler( self, signum, frame ):
        self.dump()



#!/usr/bin/env python
"""
Split large file into multiple pieces for upload to S3.
This parallelizes the task over available cores using multiprocessing.
Code mostly taken form CloudBioLinux.
"""

import contextlib
import functools
import glob
import multiprocessing
import os
import subprocess

from multiprocessing.pool import IMapIterator

try:
    import boto
    from boto.s3.connection import S3Connection
except ImportError:
    boto = None


def map_wrap(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)
    return wrapper


def mp_from_ids(s3server, mp_id, mp_keyname, mp_bucketname):
    """Get the multipart upload from the bucket and multipart IDs.

    This allows us to reconstitute a connection to the upload
    from within multiprocessing functions.
    """
    if s3server['host']:
        conn = boto.connect_s3(aws_access_key_id=s3server['access_key'],
                               aws_secret_access_key=s3server['secret_key'],
                               is_secure=s3server['is_secure'],
                               host=s3server['host'],
                               port=s3server['port'],
                               calling_format=boto.s3.connection.OrdinaryCallingFormat(),
                               path=s3server['conn_path'])
    else:
        conn = S3Connection(s3server['access_key'], s3server['secret_key'])

    bucket = conn.lookup(mp_bucketname)
    mp = boto.s3.multipart.MultiPartUpload(bucket)
    mp.key_name = mp_keyname
    mp.id = mp_id
    return mp


@map_wrap
def transfer_part(s3server, mp_id, mp_keyname, mp_bucketname, i, part):
    """Transfer a part of a multipart upload. Designed to be run in parallel.
    """
    mp = mp_from_ids(s3server, mp_id, mp_keyname, mp_bucketname)
    with open(part) as t_handle:
        mp.upload_part_from_file(t_handle, i + 1)
    os.remove(part)


def multipart_upload(s3server, bucket, s3_key_name, tarball, mb_size):
    """Upload large files using Amazon's multipart upload functionality.
    """
    cores = multiprocessing.cpu_count()

    def split_file(in_file, mb_size, split_num=5):
        prefix = os.path.join(os.path.dirname(in_file),
                              "%sS3PART" % (os.path.basename(s3_key_name)))
        max_chunk = s3server['max_chunk_size']
        # Split chunks so they are 5MB < chunk < 250MB(max_chunk_size)
        split_size = int(max(min(mb_size / (split_num * 2.0), max_chunk), 5))
        if not os.path.exists("%saa" % prefix):
            cl = ["split", "-b%sm" % split_size, in_file, prefix]
            subprocess.check_call(cl)
        return sorted(glob.glob("%s*" % prefix))

    mp = bucket.initiate_multipart_upload(s3_key_name,
                                          reduced_redundancy=s3server['use_rr'])

    with multimap(cores) as pmap:
        for _ in pmap(transfer_part, ((s3server, mp.id, mp.key_name, mp.bucket_name, i, part)
                                      for (i, part) in
                                      enumerate(split_file(tarball, mb_size, cores)))):
            pass
    mp.complete_upload()


@contextlib.contextmanager
def multimap(cores=None):
    """Provide multiprocessing imap like function.

    The context manager handles setting up the pool, worked around interrupt issues
    and terminating the pool on completion.
    """
    if cores is None:
        cores = max(multiprocessing.cpu_count() - 1, 1)

    def wrapper(func):
        def wrap(self, timeout=None):
            return func(self, timeout=timeout if timeout is not None else 1e100)
        return wrap
    IMapIterator.next = wrapper(IMapIterator.next)
    pool = multiprocessing.Pool(cores)
    yield pool.imap
    pool.terminate()



import datetime
import uuid


class Dictifiable:
    """ Mixin that enables objects to be converted to dictionaries. This is useful
        when for sharing objects across boundaries, such as the API, tool scripts,
        and JavaScript code. """

    def to_dict( self, view='collection', value_mapper=None ):
        """
        Return item dictionary.
        """

        if not value_mapper:
            value_mapper = {}

        def get_value( key, item ):
            """
            Recursive helper function to get item values.
            """
            # FIXME: why use exception here? Why not look for key in value_mapper
            # first and then default to to_dict?
            try:
                return item.to_dict( view=view, value_mapper=value_mapper )
            except:
                if key in value_mapper:
                    return value_mapper.get( key )( item )
                if type(item) == datetime.datetime:
                    return item.isoformat()
                elif type(item) == uuid.UUID:
                    return str(item)
                # Leaving this for future reference, though we may want a more
                # generic way to handle special type mappings going forward.
                # If the item is of a class that needs to be 'stringified' before being put into a JSON data structure
                # elif type(item) in []:
                #    return str(item)
                return item

        # Create dict to represent item.
        rval = dict(
            model_class=self.__class__.__name__
        )

        # Fill item dict with visible keys.
        try:
            visible_keys = self.__getattribute__( 'dict_' + view + '_visible_keys' )
        except AttributeError:
            raise Exception( 'Unknown Dictifiable view: %s' % view )
        for key in visible_keys:
            try:
                item = self.__getattribute__( key )
                if isinstance( item, list ):
                    rval[ key ] = []
                    for i in item:
                        rval[ key ].append( get_value( key, i ) )
                else:
                    rval[ key ] = get_value( key, item )
            except AttributeError:
                rval[ key ] = None

        return rval



from os.path import abspath, exists, join

from .galaxy_packages import BaseGalaxyPackageDependencyResolver, ToolShedDependency
from .resolver_mixins import UsesInstalledRepositoriesMixin
from ..resolvers import NullDependency


class ToolShedPackageDependencyResolver(BaseGalaxyPackageDependencyResolver, UsesInstalledRepositoriesMixin):
    resolver_type = "tool_shed_packages"
    # Resolution of these dependencies depends on more than just the requirement
    # tag, it depends on the tool installation context - therefore these are
    # non-simple.
    dependency_type = ToolShedDependency
    resolves_simple_dependencies = False

    def __init__(self, dependency_manager, **kwds):
        super(ToolShedPackageDependencyResolver, self).__init__(dependency_manager, **kwds)

    def _find_dep_versioned( self, name, version, type='package', **kwds ):
        installed_tool_dependency = self._get_installed_dependency( name, type, version=version, **kwds )
        if installed_tool_dependency:
            path = self._get_package_installed_dependency_path( installed_tool_dependency, name, version )
            return self._galaxy_package_dep(path, version, name, True)
        else:
            return NullDependency(version=version, name=name)

    def _find_dep_default( self, name, type='package', **kwds ):
        if type == 'set_environment' and kwds.get('installed_tool_dependencies', None):
            installed_tool_dependency = self._get_installed_dependency( name, type, version=None, **kwds )
            if installed_tool_dependency:
                dependency = self._get_set_environment_installed_dependency_script_path( installed_tool_dependency, name )
                is_galaxy_dep = isinstance(dependency, ToolShedDependency)
                has_script_dep = is_galaxy_dep and dependency.script and dependency.path
                if has_script_dep:
                    # Environment settings do not use versions.
                    return ToolShedDependency(dependency.script, dependency.path, None, name, True)
        return NullDependency(version=None, name=name)

    def _get_package_installed_dependency_path( self, installed_tool_dependency, name, version ):
        tool_shed_repository = installed_tool_dependency.tool_shed_repository
        base_path = self.base_path
        return join(
            base_path,
            name,
            version,
            tool_shed_repository.owner,
            tool_shed_repository.name,
            tool_shed_repository.installed_changeset_revision
        )

    def _get_set_environment_installed_dependency_script_path( self, installed_tool_dependency, name ):
        tool_shed_repository = installed_tool_dependency.tool_shed_repository
        base_path = self.base_path
        path = abspath( join( base_path,
                              'environment_settings',
                              name,
                              tool_shed_repository.owner,
                              tool_shed_repository.name,
                              tool_shed_repository.installed_changeset_revision ) )
        if exists( path ):
            script = join( path, 'env.sh' )
            return ToolShedDependency(script, path, None, name, True)
        return NullDependency(version=None, name=name)


__all__ = ('ToolShedPackageDependencyResolver', )



from .cwltool_deps import (
    needs_shell_quoting,
    shellescape,
)
from .parser import tool_proxy, workflow_proxy
from .representation import to_cwl_job, to_galaxy_parameters
from .runtime_actions import handle_outputs


__all__ = (
    'tool_proxy',
    'workflow_proxy',
    'handle_outputs',
    'to_cwl_job',
    'to_galaxy_parameters',
    'needs_shell_quoting',
    'shellescape',
)



from abc import (
    ABCMeta,
    abstractmethod
)

import six


@six.add_metaclass(ABCMeta)
class ToolLineage(object):
    """
    """

    @abstractmethod
    def get_versions( self, reverse=False ):
        """ Return an ordered list of lineages (ToolLineageVersion) in this
        chain, from oldest to newest.
        """


class ToolLineageVersion(object):
    """ Represents a single tool in a lineage. If lineage is based
    around GUIDs that somehow encode the version (either using GUID
    or a simple tool id and a version). """

    def __init__(self, id, version):
        self.id = id
        self.version = version

    @staticmethod
    def from_id_and_verion( id, version ):
        assert version is not None
        return ToolLineageVersion( id, version )

    @staticmethod
    def from_guid( guid ):
        return ToolLineageVersion( guid, None )

    @property
    def id_based( self ):
        """ Return True if the lineage is defined by GUIDs (in this
        case the indexer of the tools (i.e. the ToolBox) should ignore
        the tool_version (because it is encoded in the GUID and managed
        externally).
        """
        return self.version is None

    def to_dict(self):
        return dict(
            id=self.id,
            version=self.version,
        )



import bz2
import gzip
import re
import zipfile

from six import StringIO

from galaxy import util
from galaxy.util.image_util import image_type

HTML_CHECK_LINES = 100


def check_html( file_path, chunk=None ):
    if chunk is None:
        temp = open( file_path, "U" )
    else:
        temp = chunk
    regexp1 = re.compile( "<A\s+[^>]*HREF[^>]+>", re.I )
    regexp2 = re.compile( "<IFRAME[^>]*>", re.I )
    regexp3 = re.compile( "<FRAMESET[^>]*>", re.I )
    regexp4 = re.compile( "<META[\W][^>]*>", re.I )
    regexp5 = re.compile( "<SCRIPT[^>]*>", re.I )
    lineno = 0
    # TODO: Potentially reading huge lines into string here, this should be
    # reworked.
    for line in temp:
        lineno += 1
        matches = regexp1.search( line ) or regexp2.search( line ) or regexp3.search( line ) or regexp4.search( line ) or regexp5.search( line )
        if matches:
            if chunk is None:
                temp.close()
            return True
        if HTML_CHECK_LINES and (lineno > HTML_CHECK_LINES):
            break
    if chunk is None:
        temp.close()
    return False


def check_binary( name, file_path=True ):
    # Handles files if file_path is True or text if file_path is False
    is_binary = False
    if file_path:
        temp = open( name, "U" )
    else:
        temp = StringIO( name )
    try:
        for char in temp.read( 100 ):
            if util.is_binary( char ):
                is_binary = True
                break
    finally:
        temp.close( )
    return is_binary


def check_gzip( file_path ):
    # This method returns a tuple of booleans representing ( is_gzipped, is_valid )
    # Make sure we have a gzipped file
    try:
        temp = open( file_path, "U" )
        magic_check = temp.read( 2 )
        temp.close()
        if magic_check != util.gzip_magic:
            return ( False, False )
    except:
        return ( False, False )
    # We support some binary data types, so check if the compressed binary file is valid
    # If the file is Bam, it should already have been detected as such, so we'll just check
    # for sff format.
    try:
        header = gzip.open( file_path ).read(4)
        if header == b'.sff':
            return ( True, True )
    except:
        return( False, False )
    CHUNK_SIZE = 2 ** 15  # 32Kb
    gzipped_file = gzip.GzipFile( file_path, mode='rb' )
    chunk = gzipped_file.read( CHUNK_SIZE )
    gzipped_file.close()
    # See if we have a compressed HTML file
    if check_html( file_path, chunk=chunk ):
        return ( True, False )
    return ( True, True )


def check_bz2( file_path ):
    try:
        temp = open( file_path, "U" )
        magic_check = temp.read( 3 )
        temp.close()
        if magic_check != util.bz2_magic:
            return ( False, False )
    except:
        return( False, False )
    CHUNK_SIZE = 2 ** 15  # reKb
    bzipped_file = bz2.BZ2File( file_path, mode='rb' )
    chunk = bzipped_file.read( CHUNK_SIZE )
    bzipped_file.close()
    # See if we have a compressed HTML file
    if check_html( file_path, chunk=chunk ):
        return ( True, False )
    return ( True, True )


def check_zip( file_path ):
    if zipfile.is_zipfile( file_path ):
        return True
    return False


def is_bz2( file_path ):
    is_bz2, is_valid = check_bz2( file_path )
    return is_bz2


def is_gzip( file_path ):
    is_gzipped, is_valid = check_gzip( file_path )
    return is_gzipped


def check_image( file_path ):
    """ Simple wrapper around image_type to yield a True/False verdict """
    if image_type( file_path ):
        return True
    return False


__all__ = (
    'check_binary',
    'check_bz2',
    'check_gzip',
    'check_html',
    'check_image',
    'check_zip',
    'is_gzip',
    'is_bz2',
)



"""API for this module containing functionality related to the toolbox."""

from .base import AbstractToolBox, BaseGalaxyToolBox

from .panel import panel_item_types, ToolSection, ToolSectionLabel

__all__ = (
    "AbstractToolBox",
    "BaseGalaxyToolBox",
    "panel_item_types",
    "ToolSection",
    "ToolSectionLabel",
)



import bz2
import gzip
import zipfile

from .checkers import (
    is_bz2,
    is_gzip
)


def get_fileobj(filename, mode="r", gzip_only=False, bz2_only=False, zip_only=False):
    """
    Returns a fileobj. If the file is compressed, return appropriate file reader.

    :param filename: path to file that should be opened
    :param mode: mode to pass to opener
    :param gzip_only: only open file if file is gzip compressed or not compressed
    :param bz2_only: only open file if file is bz2 compressed or not compressed
    :param zip_only: only open file if file is zip compressed or not compressed
    """
    # the various compression readers don't support 'U' mode,
    # so we open in 'r'.
    if mode == 'U':
        cmode = 'r'
    else:
        cmode = mode
    if not bz2_only and not zip_only and is_gzip(filename):
        return gzip.GzipFile(filename, cmode)
    if not gzip_only and not zip_only and is_bz2(filename):
        return bz2.BZ2File(filename, cmode)
    if not bz2_only and not gzip_only and zipfile.is_zipfile(filename):
        return zipfile.os_zipfile(filename, cmode)
    return open(filename, mode)



#!/usr/bin/env python

# Copyright (c) 2006 Bermi Ferrer Martinez
#
# bermi a-t bermilabs - com
# See the end of this file for the free software, open source license (BSD-style).

import re


class Base:
    '''Locale inflectors must inherit from this base class inorder to provide
    the basic Inflector functionality'''

    def cond_plural(self, number_of_records, word):
        '''Returns the plural form of a word if first parameter is greater than 1'''

        if number_of_records != 1:
            return self.pluralize(word)
        else:
            return word

    def titleize(self, word, uppercase=''):
        '''Converts an underscored or CamelCase word into a English sentence.
            The titleize function converts text like "WelcomePage",
            "welcome_page" or  "welcome page" to this "Welcome Page".
            If second parameter is set to 'first' it will only
            capitalize the first character of the title.'''

        if(uppercase == 'first'):
            return self.humanize(self.underscore(word)).capitalize()
        else:
            return self.humanize(self.underscore(word)).title()

    def camelize(self, word):
        ''' Returns given word as CamelCased
        Converts a word like "send_email" to "SendEmail". It
        will remove non alphanumeric character from the word, so
        "who's online" will be converted to "WhoSOnline"'''
        return ''.join(w[0].upper() + w[1:] for w in re.sub('[^A-Z^a-z^0-9^:]+', ' ', word).split(' '))

    def underscore(self, word):
        ''' Converts a word "into_it_s_underscored_version"
        Convert any "CamelCased" or "ordinary Word" into an
        "underscored_word".
        This can be really useful for creating friendly URLs.'''

        return re.sub('[^A-Z^a-z^0-9^\/]+', '_',
                      re.sub('([a-z\d])([A-Z])', '\\1_\\2',
                             re.sub('([A-Z]+)([A-Z][a-z])', '\\1_\\2', re.sub('::', '/', word)))).lower()

    def humanize(self, word, uppercase=''):
        '''Returns a human-readable string from word
        Returns a human-readable string from word, by replacing
        underscores with a space, and by upper-casing the initial
        character by default.
        If you need to uppercase all the words you just have to
        pass 'all' as a second parameter.'''

        if(uppercase == 'first'):
            return re.sub('_id$', '', word).replace('_', ' ').capitalize()
        else:
            return re.sub('_id$', '', word).replace('_', ' ').title()

    def variablize(self, word):
        '''Same as camelize but first char is lowercased
        Converts a word like "send_email" to "sendEmail". It
        will remove non alphanumeric character from the word, so
        "who's online" will be converted to "whoSOnline"'''
        word = self.camelize(word)
        return word[0].lower() + word[1:]

    def tableize(self, class_name):
        ''' Converts a class name to its table name according to rails
        naming conventions. Example. Converts "Person" to "people" '''
        return self.pluralize(self.underscore(class_name))

    def classify(self, table_name):
        '''Converts a table name to its class name according to rails
        naming conventions. Example: Converts "people" to "Person" '''
        return self.camelize(self.singularize(table_name))

    def ordinalize(self, number):
        '''Converts number to its ordinal English form.
        This method converts 13 to 13th, 2 to 2nd ...'''
        tail = 'th'
        if number % 100 == 11 or number % 100 == 12 or number % 100 == 13:
            tail = 'th'
        elif number % 10 == 1:
            tail = 'st'
        elif number % 10 == 2:
            tail = 'nd'
        elif number % 10 == 3:
            tail = 'rd'

        return str(number) + tail

    def unaccent(self, text):
        '''Transforms a string to its unaccented version.
        This might be useful for generating "friendly" URLs'''
        find = u'\u00C0\u00C1\u00C2\u00C3\u00C4\u00C5\u00C6\u00C7\u00C8\u00C9\u00CA\u00CB\u00CC\u00CD\u00CE\u00CF\u00D0\u00D1\u00D2\u00D3\u00D4\u00D5\u00D6\u00D8\u00D9\u00DA\u00DB\u00DC\u00DD\u00DE\u00DF\u00E0\u00E1\u00E2\u00E3\u00E4\u00E5\u00E6\u00E7\u00E8\u00E9\u00EA\u00EB\u00EC\u00ED\u00EE\u00EF\u00F0\u00F1\u00F2\u00F3\u00F4\u00F5\u00F6\u00F8\u00F9\u00FA\u00FB\u00FC\u00FD\u00FE\u00FF'
        replace = u'AAAAAAACEEEEIIIIDNOOOOOOUUUUYTsaaaaaaaceeeeiiiienoooooouuuuyty'
        return self.string_replace(text, find, replace)

    def string_replace(self, word, find, replace):
        '''This function returns a copy of word, translating
        all occurrences of each character in find to the
        corresponding character in replace'''
        for k in range(0, len(find)):
            word = re.sub(find[k], replace[k], word)

        return word

    def urlize(self, text):
        '''Transform a string its unaccented and underscored
        version ready to be inserted in friendly URLs'''
        return re.sub('^_|_$', '', self.underscore(self.unaccent(text)))

    def demodulize(self, module_name):
        return self.humanize(self.underscore(re.sub('^.*::', '', module_name)))

    def modulize(self, module_description):
        return self.camelize(self.singularize(module_description))

    def foreignKey(self, class_name, separate_class_name_and_id_with_underscore=1):
        ''' Returns class_name in underscored form, with "_id" tacked on at the end.
        This is for use in dealing with the database.'''
        if separate_class_name_and_id_with_underscore:
            tail = '_id'
        else:
            tail = 'id'
        return self.underscore(self.demodulize(class_name)) + tail


class English (Base):
    """
    Inflector for pluralize and singularize English nouns.

    This is the default Inflector for the Inflector obj
    """

    def pluralize(self, word):
        '''Pluralizes English nouns.'''

        rules = [
            ['(?i)(quiz)$', '\\1zes'],
            ['^(?i)(ox)$', '\\1en'],
            ['(?i)([m|l])ouse$', '\\1ice'],
            ['(?i)(matr|vert|ind)ix|ex$', '\\1ices'],
            ['(?i)(x|ch|ss|sh)$', '\\1es'],
            ['(?i)([^aeiouy]|qu)ies$', '\\1y'],
            ['(?i)([^aeiouy]|qu)y$', '\\1ies'],
            ['(?i)(hive)$', '\\1s'],
            ['(?i)(?:([^f])fe|([lr])f)$', '\\1\\2ves'],
            ['(?i)sis$', 'ses'],
            ['(?i)([ti])um$', '\\1a'],
            ['(?i)(buffal|tomat)o$', '\\1oes'],
            ['(?i)(bu)s$', '\\1ses'],
            ['(?i)(alias|status)', '\\1es'],
            ['(?i)(octop|vir)us$', '\\1i'],
            ['(?i)(ax|test)is$', '\\1es'],
            ['(?i)s$', 's'],
            ['(?i)$', 's']
        ]

        uncountable_words = ['equipment', 'information', 'rice', 'money', 'species', 'series', 'fish', 'sheep']

        irregular_words = {
            'person': 'people',
            'man': 'men',
            'child': 'children',
            'sex': 'sexes',
            'move': 'moves'
        }

        lower_cased_word = word.lower()

        for uncountable_word in uncountable_words:
            if lower_cased_word[-1 * len(uncountable_word):] == uncountable_word:
                return word

        for irregular in irregular_words.keys():
            match = re.search('(' + irregular + ')$', word, re.IGNORECASE)
            if match:
                return re.sub('(?i)' + irregular + '$', match.expand('\\1')[0] + irregular_words[irregular][1:], word)

        for rule in range(len(rules)):
            match = re.search(rules[rule][0], word, re.IGNORECASE)
            if match:
                groups = match.groups()
                for k in range(0, len(groups)):
                    if groups[k] is None:
                        rules[rule][1] = rules[rule][1].replace('\\' + str(k + 1), '')

                return re.sub(rules[rule][0], rules[rule][1], word)

        return word

    def singularize(self, word):
        '''Singularizes English nouns.'''

        rules = [
            ['(?i)(quiz)zes$', '\\1'],
            ['(?i)(matr)ices$', '\\1ix'],
            ['(?i)(vert|ind)ices$', '\\1ex'],
            ['(?i)^(ox)en', '\\1'],
            ['(?i)(alias|status)es$', '\\1'],
            ['(?i)([octop|vir])i$', '\\1us'],
            ['(?i)(cris|ax|test)es$', '\\1is'],
            ['(?i)(shoe)s$', '\\1'],
            ['(?i)(o)es$', '\\1'],
            ['(?i)(bus)es$', '\\1'],
            ['(?i)([m|l])ice$', '\\1ouse'],
            ['(?i)(x|ch|ss|sh)es$', '\\1'],
            ['(?i)(m)ovies$', '\\1ovie'],
            ['(?i)(s)eries$', '\\1eries'],
            ['(?i)([^aeiouy]|qu)ies$', '\\1y'],
            ['(?i)([lr])ves$', '\\1f'],
            ['(?i)(tive)s$', '\\1'],
            ['(?i)(hive)s$', '\\1'],
            ['(?i)([^f])ves$', '\\1fe'],
            ['(?i)(^analy)ses$', '\\1sis'],
            ['(?i)((a)naly|(b)a|(d)iagno|(p)arenthe|(p)rogno|(s)ynop|(t)he)ses$', '\\1\\2sis'],
            ['(?i)([ti])a$', '\\1um'],
            ['(?i)(n)ews$', '\\1ews'],
            ['(?i)s$', ''],
        ]

        uncountable_words = ['equipment', 'information', 'rice', 'money', 'species', 'series', 'fish', 'sheep', 'sms']

        irregular_words = {
            'people': 'person',
            'men': 'man',
            'children': 'child',
            'sexes': 'sex',
            'moves': 'move'
        }

        lower_cased_word = word.lower()
        for uncountable_word in uncountable_words:
            if lower_cased_word[-1 * len(uncountable_word):] == uncountable_word:
                return word

        for irregular in irregular_words.keys():
            match = re.search('(' + irregular + ')$', word, re.IGNORECASE)
            if match:
                return re.sub('(?i)' + irregular + '$', match.expand('\\1')[0] + irregular_words[irregular][1:], word)

        for rule in range(len(rules)):
            match = re.search(rules[rule][0], word, re.IGNORECASE)
            if match:
                groups = match.groups()
                for k in range(0, len(groups)):
                    if groups[k] is None:
                        rules[rule][1] = rules[rule][1].replace('\\' + str(k + 1), '')

                return re.sub(rules[rule][0], rules[rule][1], word)

        return word


class Inflector:
    """
    Inflector for pluralizing and singularizing nouns.

    It provides methods for helping on creating programs
    based on naming conventions like on Ruby on Rails.
    """

    def __init__( self, Inflector=English ):
        assert callable(Inflector), "Inflector should be a callable obj"
        self.Inflector = Inflector()

    def pluralize(self, word):
        '''Pluralizes nouns.'''
        return self.Inflector.pluralize(word)

    def singularize(self, word):
        '''Singularizes nouns.'''
        return self.Inflector.singularize(word)

    def cond_plural(self, number_of_records, word):
        '''Returns the plural form of a word if first parameter is greater than 1'''
        return self.Inflector.cond_plural(number_of_records, word)

    def titleize(self, word, uppercase=''):
        '''Converts an underscored or CamelCase word into a sentence.
            The titleize function converts text like "WelcomePage",
            "welcome_page" or  "welcome page" to this "Welcome Page".
            If the "uppercase" parameter is set to 'first' it will only
            capitalize the first character of the title.'''
        return self.Inflector.titleize(word, uppercase)

    def camelize(self, word):
        ''' Returns given word as CamelCased
        Converts a word like "send_email" to "SendEmail". It
        will remove non alphanumeric character from the word, so
        "who's online" will be converted to "WhoSOnline"'''
        return self.Inflector.camelize(word)

    def underscore(self, word):
        ''' Converts a word "into_it_s_underscored_version"
        Convert any "CamelCased" or "ordinary Word" into an
        "underscored_word".
        This can be really useful for creating friendly URLs.'''
        return self.Inflector.underscore(word)

    def humanize(self, word, uppercase=''):
        '''Returns a human-readable string from word
        Returns a human-readable string from word, by replacing
        underscores with a space, and by upper-casing the initial
        character by default.
        If you need to uppercase all the words you just have to
        pass 'all' as a second parameter.'''
        return self.Inflector.humanize(word, uppercase)

    def variablize(self, word):
        '''Same as camelize but first char is lowercased
        Converts a word like "send_email" to "sendEmail". It
        will remove non alphanumeric character from the word, so
        "who's online" will be converted to "whoSOnline"'''
        return self.Inflector.variablize(word)

    def tableize(self, class_name):
        ''' Converts a class name to its table name according to rails
        naming conventions. Example. Converts "Person" to "people" '''
        return self.Inflector.tableize(class_name)

    def classify(self, table_name):
        '''Converts a table name to its class name according to rails
        naming conventions. Example: Converts "people" to "Person" '''
        return self.Inflector.classify(table_name)

    def ordinalize(self, number):
        '''Converts number to its ordinal form.
        This method converts 13 to 13th, 2 to 2nd ...'''
        return self.Inflector.ordinalize(number)

    def unaccent(self, text):
        '''Transforms a string to its unaccented version.
        This might be useful for generating "friendly" URLs'''
        return self.Inflector.unaccent(text)

    def urlize(self, text):
        '''Transform a string to its unaccented and underscored
        version ready to be inserted in friendly URLs'''
        return self.Inflector.urlize(text)

    def demodulize(self, module_name):
        return self.Inflector.demodulize(module_name)

    def modulize(self, module_description):
        return self.Inflector.modulize(module_description)

    def foreignKey(self, class_name, separate_class_name_and_id_with_underscore=1):
        ''' Returns class_name in underscored form, with "_id" tacked on at the end.
        This is for use in dealing with the database.'''
        return self.Inflector.foreignKey(class_name, separate_class_name_and_id_with_underscore)


# Copyright (c) 2006 Bermi Ferrer Martinez
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software to deal in this software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of this software, and to permit
# persons to whom this software is furnished to do so, subject to the following
# condition:
#
# THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THIS SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THIS SOFTWARE.



"""
Object Store plugin for the Microsoft Azure Block Blob Storage system
"""

import logging
import os
import shutil
import threading
import time

from datetime import datetime

from galaxy.exceptions import ObjectInvalid, ObjectNotFound
from galaxy.util import directory_hash_id, safe_relpath, umask_fix_perms
from galaxy.util.sleeper import Sleeper
from ..objectstore import convert_bytes, ObjectStore

try:
    from azure.common import AzureHttpError
    from azure.storage import CloudStorageAccount
    from azure.storage.blob import BlockBlobService
    from azure.storage.blob.models import Blob
except ImportError:
    BlockBlobService = None

NO_BLOBSERVICE_ERROR_MESSAGE = ("ObjectStore configured, but no azure.storage.blob dependency available."
                                "Please install and properly configure azure.storage.blob or modify Object Store configuration.")

log = logging.getLogger( __name__ )


class AzureBlobObjectStore(ObjectStore):
    """
    Object store that stores objects as blobs in an Azure Blob Container. A local
    cache exists that is used as an intermediate location for files between
    Galaxy and Azure.
    """
    def __init__(self, config, config_xml):
        if BlockBlobService is None:
            raise Exception(NO_BLOBSERVICE_ERROR_MESSAGE)
        super(AzureBlobObjectStore, self).__init__(config)

        self.staging_path = self.config.file_path
        self.transfer_progress = 0
        self._parse_config_xml(config_xml)
        self._configure_connection()
        self.container_lease = self._get_container_lease()

        # Clean cache only if value is set in galaxy.ini
        if self.cache_size != -1:
            # Convert GBs to bytes for comparison
            self.cache_size = self.cache_size * 1073741824
            # Helper for interruptable sleep
            self.sleeper = Sleeper()
            self.cache_monitor_thread = threading.Thread(target=self.__cache_monitor)
            self.cache_monitor_thread.start()
            log.info("Cache cleaner manager started")

    ###################
    # Private Methods #
    ###################

    # config_xml is an ElementTree object.
    def _parse_config_xml(self, config_xml):
        try:
            auth_xml = config_xml.find('auth')
            self.account_name = auth_xml.get('account_name')
            self.account_key = auth_xml.get('account_key')
            container_xml = config_xml.find('container')
            self.container_name = container_xml.get('name')
            self.max_chunk_size = int(container_xml.get('max_chunk_size', 250))  # currently unused
            cache_xml = config_xml.find('cache')
            self.cache_size = float(cache_xml.get('size', -1))
            self.staging_path = cache_xml.get('path', self.config.object_store_cache_path)

            for d_xml in config_xml.findall('extra_dir'):
                self.extra_dirs[d_xml.get('type')] = d_xml.get('path')

            log.debug("Object cache dir:    %s", self.staging_path)
            log.debug("       job work dir: %s", self.extra_dirs['job_work'])

        except Exception:
            # Toss it back up after logging, we can't continue loading at this point.
            log.exception("Malformed ObjectStore Configuration XML -- unable to continue")
            raise

    def _configure_connection(self):
        log.debug("Configuring Connection")
        self.account = CloudStorageAccount(self.account_name, self.account_key)
        self.service = self.account.create_block_blob_service()

    def _get_container_lease(self):
        """ Sometimes a handle to a container is not established right away so try
        it a few times. Raise error is connection is not established. """
        for i in range(5):
            try:
                self.service.break_container_lease(self.container_name)
                container_lease = self.service.acquire_container_lease(self.container_name)
                log.debug("Using azure blob store with container '%s'", self.container_name)
                return container_lease
            except AzureHttpError:
                try:
                    log.debug("container not found, creating azure blob store container with name '%s'", self.container_name)
                    self.service.create_container(self.container_name)
                    container_lease = self.service.acquire_container_lease(self.container_name)
                    return container_lease
                except AzureHttpError:
                    log.exception("Could not get container '%s', attempt %s/5", self.container_name, i + 1)
                    time.sleep(2)
        # All the attempts have been exhausted and connection was not established,
        # raise error
        raise AzureHttpError

    def _construct_path(self, obj, base_dir=None, dir_only=None, extra_dir=None, extra_dir_at_root=False, alt_name=None, obj_dir=False, **kwargs):
        # extra_dir should never be constructed from provided data but just
        # make sure there are no shenannigans afoot
        if extra_dir and extra_dir != os.path.normpath(extra_dir):
            log.warning('extra_dir is not normalized: %s', extra_dir)
            raise ObjectInvalid("The requested object is invalid")
        # ensure that any parent directory references in alt_name would not
        # result in a path not contained in the directory path constructed here
        if alt_name:
            if not safe_relpath(alt_name):
                log.warning('alt_name would locate path outside dir: %s', alt_name)
                raise ObjectInvalid("The requested object is invalid")
            # alt_name can contain parent directory references, but S3 will not
            # follow them, so if they are valid we normalize them out
            alt_name = os.path.normpath(alt_name)

        rel_path = os.path.join(*directory_hash_id(obj.id))

        if extra_dir is not None:
            if extra_dir_at_root:
                rel_path = os.path.join(extra_dir, rel_path)
            else:
                rel_path = os.path.join(rel_path, extra_dir)

        # for JOB_WORK directory
        if obj_dir:
            rel_path = os.path.join(rel_path, str(obj.id))
        if base_dir:
            base = self.extra_dirs.get(base_dir)
            return os.path.join(base, rel_path)

        # S3 folders are marked by having trailing '/' so add it now
        # rel_path = '%s/' % rel_path # assume for now we don't need this in Azure blob storage.

        if not dir_only:
            rel_path = os.path.join(rel_path, alt_name if alt_name else "dataset_%s.dat" % obj.id)

        return rel_path

    def _fix_permissions(self, rel_path):
        """ Set permissions on rel_path"""
        for basedir, _, files in os.walk(rel_path):
            umask_fix_perms(basedir, self.config.umask, 0o777, self.config.gid)
            for filename in files:
                path = os.path.join(basedir, filename)
                # Ignore symlinks
                if os.path.islink(path):
                    continue
                umask_fix_perms(path, self.config.umask, 0o666, self.config.gid)

    def _get_cache_path(self, rel_path):
        return os.path.abspath(os.path.join(self.staging_path, rel_path))

    def _get_transfer_progress(self):
        return self.transfer_progress

    def _get_size_in_azure(self, rel_path):
        try:
            properties = self.service.get_blob_properties(self.container_name, rel_path)
            # Currently this returns a blob and not a BlobProperties object
            # Similar issue for the ruby https://github.com/Azure/azure-storage-ruby/issues/13
            # The typecheck is an attempt at future-proofing this when/if the bug is fixed.
            if type(properties) is Blob:
                properties = properties.properties
            if properties:
                size_in_bytes = properties.content_length
                return size_in_bytes
        except AzureHttpError:
            log.exception("Could not get size of blob '%s' from Azure", rel_path)
            return -1

    def _in_azure(self, rel_path):
        try:
            exists = self.service.exists(self.container_name, rel_path)
        except AzureHttpError:
            log.exception("Trouble checking existence of Azure blob '%s'", rel_path)
            return False
        return exists

    def _in_cache(self, rel_path):
        """ Check if the given dataset is in the local cache. """
        cache_path = self._get_cache_path(rel_path)
        return os.path.exists(cache_path)

    def _pull_into_cache(self, rel_path):
        # Ensure the cache directory structure exists (e.g., dataset_#_files/)
        rel_path_dir = os.path.dirname(rel_path)
        if not os.path.exists(self._get_cache_path(rel_path_dir)):
            os.makedirs(self._get_cache_path(rel_path_dir))
        # Now pull in the file
        file_ok = self._download(rel_path)
        self._fix_permissions(self._get_cache_path(rel_path_dir))
        return file_ok

    def _transfer_cb(self, complete, total):
        self.transfer_progress = float(complete) / float(total) * 100  # in percent

    def _download(self, rel_path):
        local_destination = self._get_cache_path(rel_path)
        try:
            log.debug("Pulling '%s' into cache to %s", rel_path, local_destination)
            if self.cache_size > 0 and self._get_size_in_azure(rel_path) > self.cache_size:
                log.critical("File %s is larger (%s) than the cache size (%s). Cannot download.",
                             rel_path, self._get_size_in_azure(rel_path), self.cache_size)
                return False
            else:
                self.transfer_progress = 0  # Reset transfer progress counter
                self.service.get_blob_to_path(self.container_name, rel_path, local_destination, progress_callback=self._transfer_cb)
                return True
        except AzureHttpError:
            log.exception("Problem downloading '%s' from Azure", rel_path)
        return False

    def _push_to_os(self, rel_path, source_file=None, from_string=None):
        """
        Push the file pointed to by ``rel_path`` to the object store naming the blob
        ``rel_path``. If ``source_file`` is provided, push that file instead while
        still using ``rel_path`` as the blob name.
        If ``from_string`` is provided, set contents of the file to the value of
        the string.
        """
        try:
            source_file = source_file or self._get_cache_path(rel_path)

            if not os.path.exists(source_file):
                log.error("Tried updating blob '%s' from source file '%s', but source file does not exist.", rel_path, source_file)
                return False

            if os.path.getsize(source_file) == 0:
                log.debug("Wanted to push file '%s' to azure blob '%s' but its size is 0; skipping.", source_file, rel_path)
                return True

            if from_string:
                self.service.create_blob_from_text(self.container_name, rel_path, from_string, progress_callback=self._transfer_cb)
                log.debug("Pushed data from string '%s' to blob '%s'", from_string, rel_path)
            else:
                start_time = datetime.now()
                log.debug("Pushing cache file '%s' of size %s bytes to '%s'", source_file, os.path.getsize(source_file), rel_path)
                self.transfer_progress = 0  # Reset transfer progress counter
                self.service.create_blob_from_path(self.container_name, rel_path, source_file, progress_callback=self._transfer_cb)
                end_time = datetime.now()
                log.debug("Pushed cache file '%s' to blob '%s' (%s bytes transfered in %s sec)",
                          source_file, rel_path, os.path.getsize(source_file), end_time - start_time)
            return True

        except AzureHttpError:
            log.exception("Trouble pushing to Azure Blob '%s' from file '%s'", rel_path, source_file)
        return False

    ##################
    # Public Methods #
    ##################

    def exists(self, obj, **kwargs):
        in_cache = in_azure = False
        rel_path = self._construct_path(obj, **kwargs)

        in_cache = self._in_cache(rel_path)
        in_azure = self._in_azure(rel_path)
        # log.debug("~~~~~~ File '%s' exists in cache: %s; in azure: %s" % (rel_path, in_cache, in_azure))
        # dir_only does not get synced so shortcut the decision
        dir_only = kwargs.get('dir_only', False)
        base_dir = kwargs.get('base_dir', None)
        if dir_only:
            if in_cache or in_azure:
                return True
            # for JOB_WORK directory
            elif base_dir:
                if not os.path.exists(rel_path):
                    os.makedirs(rel_path)
                return True
            else:
                return False

        # TODO: Sync should probably not be done here. Add this to an async upload stack?
        if in_cache and not in_azure:
            self._push_to_os(rel_path, source_file=self._get_cache_path(rel_path))
            return True
        elif in_azure:
            return True
        else:
            return False

    def file_ready(self, obj, **kwargs):
        """
        A helper method that checks if a file corresponding to a dataset is
        ready and available to be used. Return ``True`` if so, ``False`` otherwise.
        """
        rel_path = self._construct_path(obj, **kwargs)
        # Make sure the size in cache is available in its entirety
        if self._in_cache(rel_path):
            local_size = os.path.getsize(self._get_cache_path(rel_path))
            remote_size = self._get_size_in_azure(rel_path)
            if local_size == remote_size:
                return True
            else:
                log.debug("Waiting for dataset %s to transfer from OS: %s/%s", rel_path, local_size, remote_size)

        return False

    def create(self, obj, **kwargs):

        if not self.exists(obj, **kwargs):

            # Pull out locally used fields
            extra_dir = kwargs.get('extra_dir', None)
            extra_dir_at_root = kwargs.get('extra_dir_at_root', False)
            dir_only = kwargs.get('dir_only', False)
            alt_name = kwargs.get('alt_name', None)

            # Construct hashed path
            rel_path = os.path.join(*directory_hash_id(obj.id))

            # Optionally append extra_dir
            if extra_dir is not None:
                if extra_dir_at_root:
                    rel_path = os.path.join(extra_dir, rel_path)
                else:
                    rel_path = os.path.join(rel_path, extra_dir)

            # Create given directory in cache
            cache_dir = os.path.join(self.staging_path, rel_path)
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir)

            # Although not really necessary to create S3 folders (because S3 has
            # flat namespace), do so for consistency with the regular file system
            # S3 folders are marked by having trailing '/' so add it now
            # s3_dir = '%s/' % rel_path
            # self._push_to_os(s3_dir, from_string='')
            # If instructed, create the dataset in cache & in S3
            if not dir_only:
                rel_path = os.path.join(rel_path, alt_name if alt_name else "dataset_%s.dat" % obj.id)
                open(os.path.join(self.staging_path, rel_path), 'w').close()
                self._push_to_os(rel_path, from_string='')

    def empty(self, obj, **kwargs):
        if self.exists(obj, **kwargs):
            return bool(self.size(obj, **kwargs) > 0)
        else:
            raise ObjectNotFound( 'objectstore.empty, object does not exist: %s, kwargs: %s' % ( str( obj ), str( kwargs ) ) )

    def size(self, obj, **kwargs):
        rel_path = self._construct_path(obj, **kwargs)
        if self._in_cache(rel_path):
            try:
                return os.path.getsize(self._get_cache_path(rel_path))
            except OSError as ex:
                log.info("Could not get size of file '%s' in local cache, will try Azure. Error: %s", rel_path, ex)
        elif self.exists(obj, **kwargs):
            return self._get_size_in_azure(rel_path)
        log.warning("Did not find dataset '%s', returning 0 for size", rel_path)
        return 0

    def delete(self, obj, entire_dir=False, **kwargs):
        rel_path = self._construct_path(obj, **kwargs)
        extra_dir = kwargs.get('extra_dir', None)
        base_dir = kwargs.get('base_dir', None)
        dir_only = kwargs.get('dir_only', False)
        obj_dir = kwargs.get('obj_dir', False)
        try:
            if base_dir and dir_only and obj_dir:
                # Remove temporary data in JOB_WORK directory
                shutil.rmtree(os.path.abspath(rel_path))
                return True

            # For the case of extra_files, because we don't have a reference to
            # individual files/blobs we need to remove the entire directory structure
            # with all the files in it. This is easy for the local file system,
            # but requires iterating through each individual blob in Azure and deleing it.
            if entire_dir and extra_dir:
                shutil.rmtree(self._get_cache_path(rel_path))
                blobs = self.service.list_blobs(self.container_name, prefix=rel_path)
                for blob in blobs:
                    log.debug("Deleting from Azure: %s", blob)
                    self.service.delete_blob(self.container_name, blob.name)
                return True
            else:
                # Delete from cache first
                os.unlink(self._get_cache_path(rel_path))
                # Delete from S3 as well
                if self._in_azure(rel_path):
                    log.debug("Deleting from Azure: %s", rel_path)
                    self.service.delete_blob(self.container_name, rel_path)
                    return True
        except AzureHttpError:
            log.exception("Could not delete blob '%s' from Azure", rel_path)
        except OSError:
            log.exception('%s delete error', self.get_filename(obj, **kwargs))
        return False

    def get_data(self, obj, start=0, count=-1, **kwargs):
        rel_path = self._construct_path(obj, **kwargs)
        # Check cache first and get file if not there
        if not self._in_cache(rel_path):
            self._pull_into_cache(rel_path)
        # Read the file content from cache
        data_file = open(self._get_cache_path(rel_path), 'r')
        data_file.seek(start)
        content = data_file.read(count)
        data_file.close()
        return content

    def get_filename(self, obj, **kwargs):
        rel_path = self._construct_path(obj, **kwargs)
        base_dir = kwargs.get('base_dir', None)
        dir_only = kwargs.get('dir_only', False)
        obj_dir = kwargs.get('obj_dir', False)

        # for JOB_WORK directory
        if base_dir and dir_only and obj_dir:
            return os.path.abspath(rel_path)

        cache_path = self._get_cache_path(rel_path)
        # S3 does not recognize directories as files so cannot check if those exist.
        # So, if checking dir only, ensure given dir exists in cache and return
        # the expected cache path.
        # dir_only = kwargs.get('dir_only', False)
        # if dir_only:
        #     if not os.path.exists(cache_path):
        #         os.makedirs(cache_path)
        #     return cache_path
        # Check if the file exists in the cache first
        if self._in_cache(rel_path):
            return cache_path
        # Check if the file exists in persistent storage and, if it does, pull it into cache
        elif self.exists(obj, **kwargs):
            if dir_only:  # Directories do not get pulled into cache
                return cache_path
            else:
                if self._pull_into_cache(rel_path):
                    return cache_path
        # For the case of retrieving a directory only, return the expected path
        # even if it does not exist.
        # if dir_only:
        #     return cache_path
        raise ObjectNotFound( 'objectstore.get_filename, no cache_path: %s, kwargs: %s' % ( str( obj ), str( kwargs ) ) )

        return cache_path  # Until the upload tool does not explicitly create the dataset, return expected path

    def update_from_file(self, obj, file_name=None, create=False, **kwargs):
        if create is True:
            self.create(obj, **kwargs)
        elif self.exists(obj, **kwargs):
            rel_path = self._construct_path(obj, **kwargs)
            # Chose whether to use the dataset file itself or an alternate file
            if file_name:
                source_file = os.path.abspath(file_name)
                # Copy into cache
                cache_file = self._get_cache_path(rel_path)
                try:
                    if source_file != cache_file:
                        # FIXME? Should this be a `move`?
                        shutil.copy2(source_file, cache_file)
                    self._fix_permissions(cache_file)
                except OSError:
                    log.exception("Trouble copying source file '%s' to cache '%s'", source_file, cache_file)
            else:
                source_file = self._get_cache_path(rel_path)

            self._push_to_os(rel_path, source_file)

        else:
            raise ObjectNotFound( 'objectstore.update_from_file, object does not exist: %s, kwargs: %s' % ( str( obj ), str( kwargs ) ) )

    def get_object_url(self, obj, **kwargs):
        if self.exists(obj, **kwargs):
            rel_path = self._construct_path(obj, **kwargs)
            try:
                url = self.service.make_blob_url(container_name=self.container_name, blob_name=rel_path)
                return url
            except AzureHttpError:
                log.exception("Trouble generating URL for dataset '%s'", rel_path)
        return None

    def get_store_usage_percent(self):
        return 0.0

    ##################
    # Secret Methods #
    ##################

    def __cache_monitor(self):
        time.sleep(2)  # Wait for things to load before starting the monitor
        while self.running:
            total_size = 0
            # Is this going to be too expensive of an operation to be done frequently?
            file_list = []
            for dirpath, _, filenames in os.walk(self.staging_path):
                for filename in filenames:
                    filepath = os.path.join(dirpath, filename)
                    file_size = os.path.getsize(filepath)
                    total_size += file_size
                    # Get the time given file was last accessed
                    last_access_time = time.localtime(os.stat(filepath)[7])
                    # Compose a tuple of the access time and the file path
                    file_tuple = last_access_time, filepath, file_size
                    file_list.append(file_tuple)
            # Sort the file list (based on access time)
            file_list.sort()
            # Initiate cleaning once within 10% of the defined cache size?
            cache_limit = self.cache_size * 0.9
            if total_size > cache_limit:
                log.info("Initiating cache cleaning: current cache size: %s; clean until smaller than: %s",
                         convert_bytes(total_size), convert_bytes(cache_limit))
                # How much to delete? If simply deleting up to the cache-10% limit,
                # is likely to be deleting frequently and may run the risk of hitting
                # the limit - maybe delete additional #%?
                # For now, delete enough to leave at least 10% of the total cache free
                delete_this_much = total_size - cache_limit
                # Keep deleting datasets from file_list until deleted_amount does not
                # exceed delete_this_much; start deleting from the front of the file list,
                # which assumes the oldest files come first on the list.
                deleted_amount = 0
                for entry in enumerate(file_list):
                    if deleted_amount < delete_this_much:
                        deleted_amount += entry[2]
                        os.remove(entry[1])
                        # Debugging code for printing deleted files' stats
                        # folder, file_name = os.path.split(f[1])
                        # file_date = time.strftime("%m/%d/%y %H:%M:%S", f[0])
                        # log.debug("%s. %-25s %s, size %s (deleted %s/%s)" \
                        #     % (i, file_name, convert_bytes(f[2]), file_date, \
                        #     convert_bytes(deleted_amount), convert_bytes(delete_this_much)))
                    else:
                        log.debug("Cache cleaning done. Total space freed: %s", convert_bytes(deleted_amount))

            self.sleeper.sleep(30)  # Test cache size every 30 seconds?



"""The module describes the ``collectl`` job metrics plugin."""
import logging
import os
import shutil

from galaxy import util

from ..collectl import (
    cli,
    processes,
    subsystems
)
from ..instrumenters import InstrumentPlugin
from ...metrics import formatting

log = logging.getLogger( __name__ )

# By default, only grab statistics for user processes (as identified by
# username).
DEFAULT_PROCFILT_ON = "username"
DEFAULT_SUBSYSTEMS = "process"
# Set to zero to flush every collection.
DEFAULT_FLUSH_INTERVAL = "0"

FORMATTED_RESOURCE_TITLES = {
    "PCT": "Percent CPU Usage",
    "RSYS": "Disk Reads",
    "WSYS": "Disk Writes",
}

EMPTY_COLLECTL_FILE_MESSAGE = "Skipping process summary due to empty file... job probably did not run long enough for collectl to gather data."


class CollectlFormatter( formatting.JobMetricFormatter ):

    def format( self, key, value ):
        if key == "pid":
            return ( "Process ID", int( value ) )
        elif key == "raw_log_path":
            return ( "Relative Path of Full Collectl Log", value )
        elif key == "process_max_AccumT":
            return ( "Job Runtime (System+User)", formatting.seconds_to_str( float( value ) ) )
        else:
            _, stat_type, resource_type = key.split( "_", 2 )
            if resource_type.startswith( "Vm"):
                value_str = "%s KB" % int( value )
            elif resource_type in [ "RSYS", "WSYS" ] and stat_type in [ "count", "max", "sum" ]:
                value_str = "%d (# system calls)" % int( value )
            else:
                value_str = str( value )
            resource_title = FORMATTED_RESOURCE_TITLES.get( resource_type, resource_type )
            return ( "%s (%s)" % ( resource_title, stat_type ), value_str )


class CollectlPlugin( InstrumentPlugin ):
    """ Run collectl along with job to capture system and/or process data
    according to specified collectl subsystems.
    """
    plugin_type = "collectl"
    formatter = CollectlFormatter()

    def __init__( self, **kwargs ):
        self.__configure_paths( kwargs )
        self.__configure_subsystems( kwargs )
        saved_logs_path = kwargs.get( "saved_logs_path", "" )
        if "app" in kwargs:
            log.debug("Found path for saved logs: %s" % saved_logs_path)
            saved_logs_path = kwargs[ "app" ].config.resolve_path( saved_logs_path )
        self.saved_logs_path = saved_logs_path
        self.__configure_collectl_recorder_args( kwargs )
        self.summarize_process_data = util.asbool( kwargs.get( "summarize_process_data", True ) )
        self.log_collectl_program_output = util.asbool( kwargs.get( "log_collectl_program_output", False ) )
        if self.summarize_process_data:
            if subsystems.get_subsystem( "process" ) not in self.subsystems:
                raise Exception( "Collectl plugin misconfigured - cannot summarize_process_data without process subsystem being enabled." )

            process_statistics = kwargs.get( "process_statistics", None )
            # None will let processes module use default set of statistics
            # defined there.
            self.process_statistics = processes.parse_process_statistics( process_statistics )

    def pre_execute_instrument( self, job_directory ):
        commands = []
        # Capture PID of process so we can walk its ancestors when building
        # statistics for the whole job.
        commands.append( '''echo "$$" > '%s' ''' % self.__pid_file( job_directory ) )
        # Run collectl in record mode to capture process and system level
        # statistics according to supplied subsystems.
        commands.append( self.__collectl_record_command( job_directory ) )
        return commands

    def post_execute_instrument( self, job_directory ):
        commands = []
        # collectl dies when job script completes, perhaps capture pid of
        # collectl above and check if it is still alive to allow tracking if
        # collectl ran successfully through the whole job.
        return commands

    def job_properties( self, job_id, job_directory ):
        pid = open( self.__pid_file( job_directory ), "r" ).read().strip()
        contents = os.listdir( job_directory )
        try:
            rel_path = filter( self._is_instrumented_collectl_log, contents )[ 0 ]
            path = os.path.join( job_directory, rel_path )
        except IndexError:
            message = "Failed to find collectl log in directory %s, files were %s" % ( job_directory, contents )
            raise Exception( message )

        properties = dict(
            pid=int( pid ),
        )

        if self.saved_logs_path:
            destination_rel_dir = os.path.join( *util.directory_hash_id( job_id ) )
            destination_rel_path = os.path.join( destination_rel_dir, rel_path )
            destination_path = os.path.join( self.saved_logs_path, destination_rel_path )
            destination_dir = os.path.dirname( destination_path )
            if not os.path.isdir( destination_dir ):
                os.makedirs( destination_dir )
            shutil.copyfile( path, destination_path )
            properties[ "raw_log_path" ] = destination_rel_path

        if self.summarize_process_data:
            # Run collectl in playback and generate statistics of interest
            summary_statistics = self.__summarize_process_data( pid, path )
            for statistic, value in summary_statistics:
                properties[ "process_%s" % "_".join( statistic ) ] = value

        return properties

    def __configure_paths( self, kwargs ):
        # 95% of time I would expect collectl to just be installed with apt or
        # yum, but if it is manually installed on not on path, allow
        # configuration of explicit path - and allow path to be different
        # between galaxy job handler (local_collectl_path) and compute node
        # (remote_collectl_path).
        collectl_path = kwargs.get( "collectl_path", "collectl" )
        self.remote_collectl_path = kwargs.get( "remote_collectl_path", collectl_path )
        self.local_collectl_path = kwargs.get( "local_collectl_path", collectl_path )

    def __configure_subsystems( self, kwargs ):
        raw_subsystems_str = kwargs.get( "subsystems", DEFAULT_SUBSYSTEMS )
        raw_subsystems = util.listify( raw_subsystems_str, do_strip=True )
        self.subsystems = [ subsystems.get_subsystem(_) for _ in raw_subsystems ]

    def __configure_collectl_recorder_args( self, kwargs ):
        collectl_recorder_args = kwargs.copy()

        # Allow deployer to configure separate system and process intervals,
        # but if they specify just one - use it for both. Thinking here is this
        # plugin's most useful feature is the process level information so
        # this is likely what the deployer is attempting to configure.
        if "interval" in kwargs and "interval2" not in kwargs:
            collectl_recorder_args[ "interval2" ] = kwargs[ "interval"]

        if "flush" not in kwargs:
            collectl_recorder_args[ "flush" ] = DEFAULT_FLUSH_INTERVAL

        procfilt_on = kwargs.get( "procfilt_on", DEFAULT_PROCFILT_ON ).lower()
        # Calculate explicit arguments, rest can just be passed through from
        # constructor arguments.
        explicit_args = dict(
            collectl_path=self.remote_collectl_path,
            procfilt=procfilt_argument( procfilt_on ),
            subsystems=self.subsystems,
        )
        collectl_recorder_args.update( explicit_args )
        self.collectl_recorder_args = collectl_recorder_args

    def __summarize_process_data( self, pid, collectl_log_path ):
        playback_cli_args = dict(
            collectl_path=self.local_collectl_path,
            playback_path=collectl_log_path,
            sep="9"
        )
        if not os.stat( collectl_log_path ).st_size:
            log.debug( EMPTY_COLLECTL_FILE_MESSAGE )
            return [ ]

        playback_cli = cli.CollectlCli( **playback_cli_args )
        return processes.generate_process_statistics( playback_cli, pid, self.process_statistics )

    def __collectl_recorder_cli( self, job_directory ):
        cli_args = self.collectl_recorder_args.copy()
        cli_args[ "destination_path" ] = self._instrument_file_path( job_directory, "log" )
        return cli.CollectlCli( **cli_args )

    def __collectl_record_command( self, job_directory ):
        collectl_cli = self.__collectl_recorder_cli( job_directory )
        if self.log_collectl_program_output:
            redirect_to = self._instrument_file_path( job_directory, "program_output" )
        else:
            redirect_to = "/dev/null"
        return "%s > %s 2>&1 &" % (
            collectl_cli.build_command_line(),
            redirect_to,
        )

    def __pid_file( self, job_directory ):
        return self._instrument_file_path( job_directory, "pid" )

    def _is_instrumented_collectl_log( self, filename ):
        prefix = self._instrument_file_name( "log" )
        return filename.startswith( prefix ) and filename.endswith( ".raw.gz" )


def procfilt_argument( procfilt_on ):
    if procfilt_on == "username":
        return "U$USER"
    elif procfilt_on == "uid":
        return "u$UID"
    else:
        # Ensure it is empty of None
        if procfilt_on or procfilt_on.lower() != "none":
            raise Exception( "Invalid procfilt_on argument encountered")
        return ""


__all__ = ( 'CollectlPlugin', )



import threading

from distutils.version import LooseVersion

from .interface import ToolLineage
from .interface import ToolLineageVersion


class StockLineage(ToolLineage):
    """ Simple tool's loaded directly from file system with lineage
    determined solely by distutil's LooseVersion naming scheme.
    """
    lineages_by_id = {}
    lock = threading.Lock()

    def __init__(self, tool_id, **kwds):
        self.tool_id = tool_id
        self.tool_versions = set()

    @staticmethod
    def from_tool( tool ):
        tool_id = tool.id
        lineages_by_id = StockLineage.lineages_by_id
        with StockLineage.lock:
            if tool_id not in lineages_by_id:
                lineages_by_id[ tool_id ] = StockLineage( tool_id )
        lineage = lineages_by_id[ tool_id ]
        lineage.register_version( tool.version )
        return lineage

    def register_version( self, tool_version ):
        assert tool_version is not None
        self.tool_versions.add( tool_version )

    def get_versions( self, reverse=False ):
        versions = [ ToolLineageVersion( self.tool_id, v ) for v in self.tool_versions ]
        # Sort using LooseVersion which defines an appropriate __cmp__
        # method for comparing tool versions.
        return sorted( versions, key=_to_loose_version, reverse=reverse )

    def to_dict(self):
        return dict(
            tool_id=self.tool_id,
            tool_versions=list(self.tool_versions),
            lineage_type='stock',
        )


def _to_loose_version( tool_lineage_version ):
    version = str( tool_lineage_version.version )
    return LooseVersion( version )



""" Module used to blend ini, environment, and explicit dictionary properties
to determine application configuration. Some hard coded defaults for Galaxy but
this should be reusable by tool shed and pulsar as well.
"""
import os
import os.path
import sys

from six import iteritems
from six.moves.configparser import ConfigParser


def find_config_file(default, old_default, explicit, cwd=None):
    if cwd is not None:
        default = os.path.join(cwd, default)
        old_default = os.path.join(cwd, old_default)
        if explicit is not None:
            explicit = os.path.join(cwd, explicit)

    if explicit:
        if os.path.exists(explicit):
            config_file = explicit
        else:
            raise Exception("Problem determining Galaxy's configuration - the specified configuration file cannot be found.")
    else:
        if not os.path.exists( default ) and os.path.exists( old_default ):
            config_file = old_default
        elif os.path.exists( default ):
            config_file = default
        else:
            config_file = default + ".sample"
    return config_file


def load_app_properties(
    kwds={},
    ini_file=None,
    ini_section="app:main",
    config_prefix="GALAXY_CONFIG_"
):
    properties = kwds.copy() if kwds else {}
    if ini_file:
        defaults = {
            'here': os.path.dirname(os.path.abspath(ini_file)),
            '__file__': os.path.abspath(ini_file)
        }
        parser = NicerConfigParser(ini_file, defaults=defaults)
        parser.optionxform = str  # Don't lower-case keys
        with open(ini_file) as f:
            parser.read_file(f)

        properties.update( dict( parser.items( ini_section ) ) )

    override_prefix = "%sOVERRIDE_" % config_prefix
    for key in os.environ:
        if key.startswith( override_prefix ):
            config_key = key[ len( override_prefix ): ].lower()
            properties[ config_key ] = os.environ[ key ]
        elif key.startswith( config_prefix ):
            config_key = key[ len( config_prefix ): ].lower()
            if config_key not in properties:
                properties[ config_key ] = os.environ[ key ]

    return properties


class NicerConfigParser(ConfigParser):

    def __init__(self, filename, *args, **kw):
        ConfigParser.__init__(self, *args, **kw)
        self.filename = filename
        if hasattr(self, '_interpolation'):
            self._interpolation = self.InterpolateWrapper(self._interpolation)

    read_file = getattr(ConfigParser, 'read_file', ConfigParser.readfp)

    def defaults(self):
        """Return the defaults, with their values interpolated (with the
        defaults dict itself)

        Mainly to support defaults using values such as %(here)s
        """
        defaults = ConfigParser.defaults(self).copy()
        for key, val in iteritems(defaults):
            defaults[key] = self.get('DEFAULT', key) or val
        return defaults

    def _interpolate(self, section, option, rawval, vars):
        # Python < 3.2
        try:
            return ConfigParser._interpolate(
                self, section, option, rawval, vars)
        except Exception:
            e = sys.exc_info()[1]
            args = list(e.args)
            args[0] = 'Error in file %s: %s' % (self.filename, e)
            e.args = tuple(args)
            e.message = args[0]
            raise

    class InterpolateWrapper(object):
        # Python >= 3.2
        def __init__(self, original):
            self._original = original

        def __getattr__(self, name):
            return getattr(self._original, name)

        def before_get(self, parser, section, option, value, defaults):
            try:
                return self._original.before_get(parser, section, option,
                                                 value, defaults)
            except Exception:
                e = sys.exc_info()[1]
                args = list(e.args)
                args[0] = 'Error in file %s: %s' % (parser.filename, e)
                e.args = tuple(args)
                e.message = args[0]
                raise


__all__ = ('find_config_file', 'load_app_properties', 'NicerConfigParser')



"""This module defines Galaxy's custom exceptions.

A Galaxy exception is an exception that extends :class:`MessageException` which
defines an HTTP status code (represented by the `status_code` attribute) and a
default error message.

New exceptions should be defined by adding an entry to `error_codes.json` in this
directory to define a default error message and a Galaxy "error code". A concrete
Python class should be added in this file defining an HTTP status code (as
`status_code`) and error code (`error_code`) object loaded dynamically from
`error_codes.json`.

Reflecting Galaxy's origins as a web application, these exceptions tend to be a
bit web-oriented. However this module is a dependency of modules and tools that
have nothing to do with the web - keep this in mind when defining exception names
and messages.
"""

from ..exceptions import error_codes


class MessageException( Exception ):
    """Most generic Galaxy exception - indicates merely that some exceptional condition happened."""
    # status code to be set when used with API.
    status_code = 400
    # Error code information embedded into API json responses.
    err_code = error_codes.UNKNOWN

    def __init__( self, err_msg=None, type="info", **extra_error_info ):
        self.err_msg = err_msg or self.err_code.default_error_message
        self.type = type
        self.extra_error_info = extra_error_info

    def __str__( self ):
        return self.err_msg


class ItemDeletionException( MessageException ):
    pass


class ObjectInvalid( Exception ):
    """ Accessed object store ID is invalid """
    pass

# Please keep the exceptions ordered by status code


class ActionInputError( MessageException ):
    status_code = 400
    err_code = error_codes.USER_REQUEST_INVALID_PARAMETER

    def __init__( self, err_msg, type="error" ):
        super( ActionInputError, self ).__init__( err_msg, type )


class DuplicatedSlugException( MessageException ):
    status_code = 400
    err_code = error_codes.USER_SLUG_DUPLICATE


class DuplicatedIdentifierException( MessageException ):
    status_code = 400
    err_code = error_codes.USER_IDENTIFIER_DUPLICATE


class ObjectAttributeInvalidException( MessageException ):
    status_code = 400
    err_code = error_codes.USER_OBJECT_ATTRIBUTE_INVALID


class ObjectAttributeMissingException( MessageException ):
    status_code = 400
    err_code = error_codes.USER_OBJECT_ATTRIBUTE_MISSING


class MalformedId( MessageException ):
    status_code = 400
    err_code = error_codes.MALFORMED_ID


class MalformedContents( MessageException ):
    status_code = 400
    err_code = error_codes.MALFORMED_CONTENTS


class UnknownContentsType( MessageException ):
    status_code = 400
    err_code = error_codes.UNKNOWN_CONTENTS_TYPE


class RequestParameterMissingException( MessageException ):
    status_code = 400
    err_code = error_codes.USER_REQUEST_MISSING_PARAMETER


class ToolMetaParameterException( MessageException ):
    status_code = 400
    err_code = error_codes.USER_TOOL_META_PARAMETER_PROBLEM


class ToolMissingException( MessageException ):
    status_code = 400
    err_code = error_codes.USER_TOOL_MISSING_PROBLEM


class RequestParameterInvalidException( MessageException ):
    status_code = 400
    err_code = error_codes.USER_REQUEST_INVALID_PARAMETER


class AuthenticationFailed( MessageException ):
    status_code = 401
    err_code = error_codes.USER_AUTHENTICATION_FAILED


class AuthenticationRequired( MessageException ):
    status_code = 403
    # TODO: as 401 and send WWW-Authenticate: ???
    err_code = error_codes.USER_NO_API_KEY


class ItemAccessibilityException( MessageException ):
    status_code = 403
    err_code = error_codes.USER_CANNOT_ACCESS_ITEM


class ItemOwnershipException( MessageException ):
    status_code = 403
    err_code = error_codes.USER_DOES_NOT_OWN_ITEM


class ConfigDoesNotAllowException( MessageException ):
    status_code = 403
    err_code = error_codes.CONFIG_DOES_NOT_ALLOW


class InsufficientPermissionsException( MessageException ):
    status_code = 403
    err_code = error_codes.INSUFFICIENT_PERMISSIONS


class AdminRequiredException( MessageException ):
    status_code = 403
    err_code = error_codes.ADMIN_REQUIRED


class ObjectNotFound( MessageException ):
    """ Accessed object was not found """
    status_code = 404
    err_code = error_codes.USER_OBJECT_NOT_FOUND


class DeprecatedMethod( MessageException ):
    """
    Method (or a particular form/arg signature) has been removed and won't be available later
    """
    status_code = 404
    # TODO:?? 410 Gone?
    err_code = error_codes.DEPRECATED_API_CALL


class Conflict( MessageException ):
    status_code = 409
    err_code = error_codes.CONFLICT


class ConfigurationError( Exception ):
    status_code = 500
    err_code = error_codes.CONFIG_ERROR


class InconsistentDatabase ( MessageException ):
    status_code = 500
    err_code = error_codes.INCONSISTENT_DATABASE


class InternalServerError ( MessageException ):
    status_code = 500
    err_code = error_codes.INTERNAL_SERVER_ERROR


class NotImplemented ( MessageException ):
    status_code = 501
    err_code = error_codes.NOT_IMPLEMENTED



import random
import shlex
import socket
import subprocess


def unused_port(range=None):
    if range:
        return __unused_port_on_range(range)
    else:
        return __unused_port_rangeless()


def __unused_port_rangeless():
    # TODO: Allow ranges (though then need to guess and check)...
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('localhost', 0))
    addr, port = s.getsockname()
    s.close()
    return port


def __unused_port_on_range(range):
    assert range[0] and range[1]

    # Find all ports that are already occupied
    cmd_netstat = shlex.split("netstat tuln")
    p1 = subprocess.Popen(cmd_netstat, stdout=subprocess.PIPE)

    occupied_ports = set()
    for line in p1.stdout.read().split('\n'):
        if line.startswith('tcp') or line.startswith('tcp6'):
            col = line.split()
            local_address = col[3]
            local_port = local_address.split(':')[1]
            occupied_ports.add( int(local_port) )

    # Generate random free port number.
    while True:
        port = random.randrange(range[0], range[1])
        if port not in occupied_ports:
            break
    return port



"""
Utility functions for bi-directional Python version compatibility.  Python 2.5
introduced hashlib which replaced sha in Python 2.4 and previous versions.
"""
import hashlib
import hmac
import logging

log = logging.getLogger( __name__ )

sha1 = hashlib.sha1
sha = sha1
md5 = hashlib.md5


def md5_hash_file(path):
    """
    Return a md5 hashdigest for a file.
    """
    hasher = hashlib.md5()
    with open(path, 'rb') as afile:
        buf = afile.read()
        hasher.update(buf)
        return hasher.hexdigest()


def new_secure_hash( text_type=None ):
    """
    Returns either a sha1 hash object (if called with no arguments), or a
    hexdigest of the sha1 hash of the argument `text_type`.
    """
    if text_type:
        return sha1( text_type ).hexdigest()
    else:
        return sha1()


def hmac_new( key, value ):
    return hmac.new( key, value, sha ).hexdigest()


def is_hashable( value ):
    try:
        hash( value )
    except:
        return False
    return True


__all__ = ('md5', 'hashlib', 'sha1', 'sha', 'new_secure_hash', 'hmac_new', 'is_hashable')



"""The module describes the ``cpuinfo`` job metrics plugin."""
import logging
import re

from galaxy import util

from ..instrumenters import InstrumentPlugin
from ...metrics import formatting

log = logging.getLogger( __name__ )

PROCESSOR_LINE = re.compile(r"processor\s*\:\s*(\d+)")


class CpuInfoFormatter( formatting.JobMetricFormatter ):

    def format( self, key, value ):
        if key == "processor_count":
            return "Processor Count", "%s" % int( value )
        else:
            return key, value


class CpuInfoPlugin( InstrumentPlugin ):
    """ Gather information about processor configuration from /proc/cpuinfo.
    Linux only.
    """
    plugin_type = "cpuinfo"
    formatter = CpuInfoFormatter()

    def __init__( self, **kwargs ):
        self.verbose = util.asbool( kwargs.get( "verbose", False ) )

    def pre_execute_instrument( self, job_directory ):
        return "cat /proc/cpuinfo > '%s'" % self.__instrument_cpuinfo_path( job_directory )

    def job_properties( self, job_id, job_directory ):
        properties = {}
        processor_count = 0
        with open( self.__instrument_cpuinfo_path( job_directory ) ) as f:
            current_processor = None
            for line in f:
                line = line.strip().lower()
                if not line:  # Skip empty lines
                    continue

                processor_line_match = PROCESSOR_LINE.match( line )
                if processor_line_match:
                    processor_count += 1
                    current_processor = processor_line_match.group( 1 )
                elif current_processor and self.verbose:
                    # If verbose, dump information about each processor
                    # into database...
                    key, value = line.split( ":", 1 )
                    key = "processor_%s_%s" % ( current_processor, key.strip() )
                    value = value
        properties[ "processor_count" ] = processor_count
        return properties

    def __instrument_cpuinfo_path( self, job_directory ):
        return self._instrument_file_path( job_directory, "cpuinfo" )


__all__ = ( 'CpuInfoPlugin', )



from abc import abstractmethod

from six import iteritems

from galaxy.util import bunch
from galaxy.util.dictifiable import Dictifiable
from galaxy.util.odict import odict

from .parser import ensure_tool_conf_item


panel_item_types = bunch.Bunch(
    TOOL="TOOL",
    WORKFLOW="WORKFLOW",
    SECTION="SECTION",
    LABEL="LABEL",
)


class HasPanelItems:
    """
    """

    @abstractmethod
    def panel_items( self ):
        """ Return an ordered dictionary-like object describing tool panel
        items (such as workflows, tools, labels, and sections).
        """

    def panel_items_iter( self ):
        """ Iterate through panel items each represented as a tuple of
        (panel_key, panel_type, panel_content).
        """
        for panel_key, panel_value in iteritems(self.panel_items()):
            if panel_value is None:
                continue
            panel_type = panel_item_types.SECTION
            if panel_key.startswith("tool_"):
                panel_type = panel_item_types.TOOL
            elif panel_key.startswith("label_"):
                panel_type = panel_item_types.LABEL
            elif panel_key.startswith("workflow_"):
                panel_type = panel_item_types.WORKFLOW
            yield (panel_key, panel_type, panel_value)


class ToolSection( Dictifiable, HasPanelItems, object ):
    """
    A group of tools with similar type/purpose that will be displayed as a
    group in the user interface.
    """

    dict_collection_visible_keys = ( 'id', 'name', 'version' )

    def __init__( self, item=None ):
        """ Build a ToolSection from an ElementTree element or a dictionary.
        """
        if item is None:
            item = dict()
        self.name = item.get('name') or ''
        self.id = item.get('id') or ''
        self.version = item.get('version') or ''
        self.elems = ToolPanelElements()

    def copy( self ):
        copy = ToolSection()
        copy.name = self.name
        copy.id = self.id
        copy.version = self.version
        copy.elems = self.elems.copy()
        return copy

    def to_dict( self, trans, link_details=False ):
        """ Return a dict that includes section's attributes. """

        section_dict = super( ToolSection, self ).to_dict()
        section_elts = []
        kwargs = dict(
            trans=trans,
            link_details=link_details
        )
        for elt in self.elems.values():
            section_elts.append( elt.to_dict( **kwargs ) )
        section_dict[ 'elems' ] = section_elts

        return section_dict

    def panel_items( self ):
        return self.elems


class ToolSectionLabel( Dictifiable, object ):
    """
    A label for a set of tools that can be displayed above groups of tools
    and sections in the user interface
    """

    dict_collection_visible_keys = ( 'id', 'text', 'version' )

    def __init__( self, item ):
        """ Build a ToolSectionLabel from an ElementTree element or a
        dictionary.
        """
        item = ensure_tool_conf_item(item)
        self.text = item.get( "text" )
        self.id = item.get( "id" )
        self.version = item.get( "version" ) or ''

    def to_dict( self, **kwds ):
        return super( ToolSectionLabel, self ).to_dict()


class ToolPanelElements( HasPanelItems, odict ):
    """ Represents an ordered dictionary of tool entries - abstraction
    used both by tool panel itself (normal and integrated) and its sections.
    """

    def update_or_append( self, index, key, value ):
        if key in self or index is None:
            self[ key ] = value
        else:
            self.insert( index, key, value )

    def has_tool_with_id( self, tool_id ):
        key = 'tool_%s' % tool_id
        return key in self

    def replace_tool( self, previous_tool_id, new_tool_id, tool ):
        previous_key = 'tool_%s' % previous_tool_id
        new_key = 'tool_%s' % new_tool_id
        index = self.keys().index( previous_key )
        del self[ previous_key ]
        self.insert( index, new_key, tool )

    def index_of_tool_id( self, tool_id ):
        query_key = 'tool_%s' % tool_id
        for index, target_key in enumerate( self.keys() ):
            if query_key == target_key:
                return index
        return None

    def insert_tool( self, index, tool ):
        key = "tool_%s" % tool.id
        self.insert( index, key, tool )

    def get_tool_with_id( self, tool_id ):
        key = "tool_%s" % tool_id
        return self[ key ]

    def append_tool( self, tool ):
        key = "tool_%s" % tool.id
        self[ key ] = tool

    def stub_tool( self, key ):
        key = "tool_%s" % key
        self[ key ] = None

    def stub_workflow( self, key ):
        key = 'workflow_%s' % key
        self[ key ] = None

    def stub_label( self, key ):
        key = 'label_%s' % key
        self[ key ] = None

    def append_section( self, key, section_elems ):
        self[ key ] = section_elems

    def panel_items( self ):
        return self
